/*
 * Copyright 2009      Luc Verhaegen.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <stdio.h>

#include <X11/Xlibint.h>
#include <X11/extensions/Xv.h>
#include <X11/extensions/Xvlib.h>
#include <X11/extensions/Xvproto.h>
#include <X11/extensions/XvMC.h>
#include <X11/extensions/XvMCproto.h>
#include <X11/extensions/XvMClib.h>
#include <X11/extensions/Xext.h>
#include <X11/extensions/extutil.h>
#include <X11/extensions/vldXvMC.h>

#include "xvmce.h"
#include "xvmce_proto.h"

/* enable this to have extra debugging */
#if 0
#define MPEGDebug(x) (x)
#else
#define MPEGDebug(x)
#endif

#define FOURCC_MPEG (('G' << 24) | ('E' << 16) | ('P' << 8) | 'M')

/* Oh you must be kidding me. What a bunch of clueless idiots.
   Who splits such an API all over the place? */
extern Status _xvmc_create_context(Display *dpy, XvMCContext *context,
				   int *priv_count, CARD32 **priv_data);

extern Status _xvmc_destroy_context(Display *dpy, XvMCContext *context);

extern Status _xvmc_create_surface(Display *dpy, XvMCContext *context,
				   XvMCSurface *surface, int *priv_count,
				   CARD32 **priv_data);

extern Status _xvmc_destroy_surface(Display *dpy, XvMCSurface *surface);

extern Status _xvmc_create_subpicture(Display *dpy, XvMCContext *context,
				      XvMCSubpicture *subpicture,
				      int *priv_count, uint **priv_data);

extern Status _xvmc_destroy_subpicture(Display *dpy,
				       XvMCSubpicture *subpicture);


static int XvMC_EventBase;
static int XvMC_ErrorBase;
static int XvMCE_EventBase;
static int XvMCE_ErrorBase;

static XID current_surface_id = 0;
static int current_surface_size = 0;

/* because of the rather abhorrent nature of the XvMC interface. */
static XvPortID XvMCPortID = 0;
static XvImage *XvMCImage = NULL;
static XID XvMCImageID = 0;
static Drawable XvMCDrawable = 0;
static GC XvMCDrawableGC = NULL;

/*
 *
 */
_X_EXPORT Status
XvMCCreateContext(Display *display, XvPortID port, int surface_type_id,
		  int width, int height, int flags, XvMCContext *context)
{
    Status status;
    int major, minor;
    int priv_count;
    CARD32 *priv_data = NULL;
    XvImageFormatValues *ImageFormats;
    int count, i;

    MPEGDebug(printf("%s\n", __func__));

    if (!display || !context)
	return BadValue;

    /* First, check XvMC. */
    if (!XvMCQueryExtension(display, &XvMC_EventBase, &XvMC_ErrorBase)) {
	printf("%s: XvMC extension is missing.\n", __func__);
	return BadRequest;
    }

    status = XvMCQueryVersion(display, &major, &minor);
    if (status != Success) {
	printf("%s: XvMCQueryVersion failed: %d\n", __func__, status);
	return BadImplementation;
    }
    printf("Found XvMC Extension version %d.%d.\n", major, minor);

    /* Now, check our own extension */
    status = XvMCEQueryExtension(display, &major, &minor,
				 &XvMCE_EventBase, &XvMCE_ErrorBase);
    if (status != Success) {
	printf("%s: XvMCEQueryExtension failed: %d\n", __func__, status);
	return status;
    }
    printf("Found Unichrome XvMCE Extension version %d.%d.\n", major, minor);

    /* Find our *special* FOURCC */
    ImageFormats = XvListImageFormats(display, port, &count);
    for (i = 0; i < count; i++)
	if (ImageFormats[i].id == FOURCC_MPEG)
	    break;

    XFree(ImageFormats); /* don't need it anymore. */

    if (i == count) {
	fprintf(stderr, "%s: Unable to find FOURCC_MPEG.\n", __func__);
	return BadImplementation;
    }

    /* Create the image for Xv to use. */
    XvMCImage = XvCreateImage(display, port, FOURCC_MPEG,
			      (char *) &XvMCImageID, width, height);
    if (!XvMCImage) {
	fprintf(stderr, "%s: XvCreateImage Failed.\n", __func__);
	return BadImplementation;
    }

    /* Only now, we can go and try to create a context, with a twist. */
    context->surface_type_id = surface_type_id;
    context->width = width; /* this better be correct, or we die */
    context->height = height;
    context->flags = flags;
    context->port = port;

    /* now call the old, raped X extension hooks */
    status =_xvmc_create_context(display, context, &priv_count, &priv_data);
    if (status != Success)
	printf("%s: _xvmc_create_context failed: %d\n", __func__, status);

    /* Ok, this seems to be valid, and we can set up the PortID as well */
    XvMCPortID = port;

    return status;
}


/*
 *
 */
_X_EXPORT Status
XvMCDestroyContext(Display *display, XvMCContext *context)
{
    Status status;

    MPEGDebug(printf("%s\n", __func__));

    if (!display || !context)
	return BadValue;

    if (XvMCDrawable) {
	status = XvStopVideo(display, XvMCPortID, XvMCDrawable);
	if (status != Success)
	    fprintf(stderr, "%s: XvStopVideo failed: %d\n", __func__, status);
    }
    XvMCDrawable = 0;

    if (XvMCImage)
	XFree(XvMCImage);
    XvMCImage = NULL;

    XvMCPortID = 0;

    status = _xvmc_destroy_context(display, context);
    if (status != Success)
	fprintf(stderr, "%s: _xvmc_destroy_context failed: %d\n", __func__, status);
    XvMCDrawable = 0;

    return Success;
}


/*
 *
 */
_X_EXPORT Status
XvMCCreateSurface(Display *display, XvMCContext *context, XvMCSurface *surface)
{
    Status status;
    int priv_count;
    CARD32 *priv_data = NULL;

    MPEGDebug(printf("%s\n", __func__));

    if (!display || !context)
        return XvMCBadContext;

    if (!surface)
	return XvMCBadSurface;

    status = _xvmc_create_surface(display, context, surface,
				  &priv_count, &priv_data);
    if (status != Success)
	printf("%s: _xvmc_create_surface failed: %d\n", __func__, status);

    MPEGDebug(printf("%s: Claimed 0x%08X.\n", __func__, surface->surface_id));

    return status;
}


/*
 *
 */
_X_EXPORT Status
XvMCDestroySurface(Display *display, XvMCSurface *surface)
{
    MPEGDebug(printf("%s\n", __func__));

    if (!display || !surface)
	return BadValue;

    MPEGDebug(printf("%s: releasing 0x%08X\n", __func__, surface->surface_id));

    return _xvmc_destroy_surface(display, surface);
}

/*
 *
 */
_X_EXPORT Status
XvMCPutSurface(Display *display, XvMCSurface *surface, Drawable draw,
	       short src_x, short src_y, unsigned short src_w,
	       unsigned short src_h, short dst_x, short dst_y,
	       unsigned short dst_w, unsigned short dst_h, int flags)
{
    Status status;

    MPEGDebug(printf("%s: 0x%08X\n", __func__, surface->surface_id));

    /* We have to fake our own GC, how cool an interface is this? */
    if (XvMCDrawable != draw) {
	if (XvMCDrawableGC)
	    XFreeGC(display, XvMCDrawableGC);
	XvMCDrawableGC = XCreateGC(display, draw, 0, NULL);
    }

    XvMCDrawable = draw;
    XvMCImageID = surface->surface_id;

    status = XvPutImage(display, XvMCPortID, XvMCDrawable, XvMCDrawableGC,
			XvMCImage, src_x, src_y, src_w, src_h,
			dst_x, dst_y, dst_w, dst_h);
    if (status != Success)
	fprintf(stderr, "%s: XvPutImage failed: %d\n", __func__, status);

    return status;
}


/*
 *
 */
_X_EXPORT Status
XvMCHideSurface(Display *display, XvMCSurface *surface)
{
    Status status = Success;

    MPEGDebug(printf("%s\n", __func__));

    if (XvMCDrawable)
	status = XvStopVideo(display, XvMCPortID, XvMCDrawable);

    if (status != Success)
	fprintf(stderr, "%s: XvStopVideo failed: %d\n", __func__, status);

    XvMCDrawable = 0;

    return status;
}


/*
 *
 */
_X_EXPORT Status
XvMCFlushSurface(Display *display, XvMCSurface *surface)
{
    MPEGDebug(printf("%s: 0x%08X (%dbytes)\n", __func__,
		     surface->surface_id, current_surface_size));

    /* We don't need to do this with our hardware. */

    return Success;
}


/*
 *
 */
_X_EXPORT Status
XvMCGetSurfaceStatus(Display *display, XvMCSurface *surface, int *stat)
{
    MPEGDebug(printf("%s: 0x%08X\n", __func__, surface->surface_id));

    /* Ah, the advantage of tracking everything serverside.
       - to check whether some */
    *stat = 0;

    return Success;
}


/*
 *
 */
_X_EXPORT Status
XvMCSyncSurface(Display *display, XvMCSurface *surface)
{
    MPEGDebug(printf("%s: 0x%08X\n", __func__, surface->surface_id));

    /* This implementation syncs right before flipping at the end of the
       XvPutImage. */

    return Success;
}


/*
 *
 */
_X_EXPORT Status
XvMCBeginSurface(Display *display, XvMCContext *context, XvMCSurface *target,
		 XvMCSurface *past, XvMCSurface *future,
		 const XvMCMpegControl *control)
{
    Status status;
    CARD8 PictureType, SliceFlags;

    if (current_surface_id == target->surface_id)
	return BadValue;

    current_surface_id = target->surface_id;
    current_surface_size = 0;

    MPEGDebug(printf("%s: 0x%08X (0x%08X <- | -> 0x%08X)\n", __func__,
		     target->surface_id, past ? past->surface_id : 0,
		     future? future->surface_id : 0));

#if 0
    printf("Backwards vertical: %d\n", control->BVMV_range);
    printf("Backwards horizontal: %d\n", control->BHMV_range);
    printf("Forwards vertical: %d\n", control->FVMV_range);
    printf("Forwards horizontal: %d\n", control->FHMV_range);

    printf("Picture Structure: %d\n", control->picture_structure);

    printf("intra_dc_precision: %d\n", control->intra_dc_precision);
    printf("picture_coding_type: %d\n", control->picture_coding_type);
    printf("mpeg_coding: %d\n", control->mpeg_coding);
    printf("flags: 0x%08X\n", control->flags);
#endif

    PictureType = 0;
    PictureType |= control->picture_structure &
	XVMCE_PICTURETYPE_STRUCTURE_MASK;
    PictureType |= (control->picture_coding_type << 2) &
	XVMCE_PICTURETYPE_FRAMETYPE_MASK;
    if (control->flags & XVMC_ALTERNATE_SCAN)
	PictureType |= XVMCE_PICTURETYPE_SCAN_ALTERNATE;

    SliceFlags = 0;
    if (control->flags & XVMC_PRED_DCT_FRAME)
	SliceFlags |= XVMCE_SLICEFLAGS_PREDICTION_FRAME;
    if (control->flags & XVMC_TOP_FIELD_FIRST)
	SliceFlags |= XVMCE_SLICEFLAGS_FIELD_FIRST_TOP;
    if (control->flags & XVMC_CONCEALMENT_MOTION_VECTORS)
	SliceFlags |= XVMCE_SLICEFLAGS_INTRA_HAS_MV;
    if (control->flags & XVMC_INTRA_VLC_FORMAT)
	SliceFlags |= XVMCE_SLICEFLAGS_INTRA_DC_TABLE_MASK;
    SliceFlags |= (control->intra_dc_precision << 4) &
	XVMCE_SLICEFLAGS_INTRA_PRECISION_MASK;
    if (control->flags & XVMC_Q_SCALE_TYPE)
	SliceFlags |= XVMCE_SLICEFLAGS_QUANTIZATION_TABLE_MASK;

    status = XVMCEBufferSliceInit(display, XvMCPortID, target->surface_id,
				  PictureType, SliceFlags,
				  past ? 1 : 0, past ? past->surface_id : 0,
				  future ? 1 : 0, future ? future->surface_id : 0,
				  control->BHMV_range, control->BVMV_range,
				  control->FHMV_range, control->FVMV_range);
    if (status != Success) {
	fprintf(stderr, "%s: XVMCEBufferSliceInit failed: %d\n",
		__func__, status);
	return status;
    }

    return Success;
}


/*
 *
 */
_X_EXPORT Status
XvMCPutSlice(Display *display, XvMCContext *context,
	     char *slice, int size)
{
    Status status;

#if 0 /* loud. */
    MPEGDebug(printf("%s:0x%08X (%dbytes): %08X\n", __func__,
		     current_surface_id, size, *((unsigned int *) slice)));
#endif

    current_surface_size += size;

    status = XVMCEBufferSliceSend(display, XvMCPortID, current_surface_id,
				  slice[3], size - 4, slice + 4);
    if (status != Success)
	fprintf(stderr, "%s: XVMCEBufferSliceSend failed: %d\n",
		__func__, status);

    return status;
}


/*
 *
 */
_X_EXPORT Status
XvMCPutSlice2(Display *display, XvMCContext *context,
	      char *slice, int size, int slicecode)
{
    Status status;

#if 0 /* loud. */
    MPEGDebug(printf("%s: 0x%08X: %02X (%dbytes)\n", __func__,
		     current_surface_id, slicecode, size));
#endif

    current_surface_size += size;

    status = XVMCEBufferSliceSend(display, XvMCPortID, current_surface_id,
				  slicecode, size, slice);
    if (status != Success)
	fprintf(stderr, "%s: XVMCEBufferSliceSend failed: %d\n",
		__func__, status);

    return status;
}


/*
 *
 */
_X_EXPORT Status
XvMCLoadQMatrix(Display *display, XvMCContext *context,
		const XvMCQMatrix *qmatrix)
{
    Status status = Success;

    if (!display || !context)
        return XvMCBadContext;

    if (!qmatrix)
	return BadValue;

    if (qmatrix->load_intra_quantiser_matrix) {
	status = XvMCEQMatrixSend(display, context->port, XVMCE_QMATRIX_INTRA,
				  qmatrix->intra_quantiser_matrix);
	if (status != Success)
	    printf("%s: XvMCEQMatrixSend %s failed: %d\n", __func__,
		   "INTRA", status);
    }

    if (qmatrix->load_non_intra_quantiser_matrix) {
	status = XvMCEQMatrixSend(display, context->port, XVMCE_QMATRIX_NONINTRA,
				  qmatrix->non_intra_quantiser_matrix);
	if (status != Success)
	    printf("%s: XvMCEQMatrixSend %s failed: %d\n", __func__,
		   "NONINTRA", status);
    }

    if (qmatrix->load_chroma_intra_quantiser_matrix) {
	status = XvMCEQMatrixSend(display, context->port, XVMCE_QMATRIX_INTRA_CHROMA,
				  qmatrix->chroma_intra_quantiser_matrix);
	if (status != Success)
	    printf("%s: XvMCEQMatrixSend %s failed: %d\n", __func__,
		   "INTRA_CHROMA", status);
    }

    if (qmatrix->load_chroma_non_intra_quantiser_matrix) {
	status = XvMCEQMatrixSend(display, context->port, XVMCE_QMATRIX_NONINTRA_CHROMA,
				  qmatrix->chroma_non_intra_quantiser_matrix);
	if (status != Success)
	    printf("%s: XvMCEQMatrixSend %s failed: %d\n", __func__,
		   "NONINTRA_CHROME", status);
    }

    /* return last failure/success */
    return status;
}


/*
 *
 * Xv Attribute system wrapping.
 *
 */

/*
 *
 */
_X_EXPORT XvAttribute *
XvMCQueryAttributes(Display *display, XvMCContext *context, int *number)
{
    return XvQueryPortAttributes(display, context->port, number);
}


/*
 *
 */
_X_EXPORT Status
XvMCSetAttribute(Display *display, XvMCContext *context,
		 Atom Attribute, int value)
{
    return XvSetPortAttribute(display, context->port, Attribute, value);
}


/*
 *
 */
_X_EXPORT Status
XvMCGetAttribute(Display *display, XvMCContext *context,
		 Atom Attribute, int *value)
{
    return XvGetPortAttribute(display, context->port, Attribute, value);
}


/*
 *
 * Subpicture support.
 *
 */

/*
 *
 * During the bringup of subpicture support it was clear that there was little
 * to no correlation with the subpicture APIs implemented on top of XvMC and
 * the "simpler overlay" inside an overlay that is the subpicture support on
 * unichrome hardware. Rather than mucking about, or rather than redesigning
 * a working implementation, why not just claim not to support subpictures, and
 * have media players overwrite the colorkey in the framebuffer directly.
 *
 * When a winner is chosen from VDPAU, VA-API and whatever, wake me up, i'll
 * be happy to port the once working "XUMA" unichrome subpicture support to
 * that.
 */

/*
 *
 */
_X_EXPORT Status
XvMCSyncSubpicture(Display *display, XvMCSubpicture *subpicture)
{
    fprintf(stderr, "%s is not implemented.\n", __func__);

    return BadImplementation;
}


/*
 *
 */
_X_EXPORT Status
XvMCFlushSubpicture(Display *display, XvMCSubpicture *subpicture)
{
    fprintf(stderr, "%s is not implemented.\n", __func__);

    return Success;
}


/*
 *
 */
_X_EXPORT Status
XvMCGetSubpictureStatus(Display *display, XvMCSubpicture *subpicture, int *stat)
{
    fprintf(stderr, "%s is not implemented.\n", __func__);

    return BadImplementation;
}


/*
 *
 */
_X_EXPORT Status
XvMCCreateSubpicture(Display *display, XvMCContext *context,
		     XvMCSubpicture *subpicture, unsigned short width,
		     unsigned short height, int xvimage_id)
{
    fprintf(stderr, "%s is not implemented.\n", __func__);

    return BadImplementation;
}


/*
 *
 */
_X_EXPORT Status
XvMCClearSubpicture(Display *display, XvMCSubpicture *subpicture,
		    short x, short y, unsigned short width,
		    unsigned short height, unsigned int colour)
{
    fprintf(stderr, "%s is not implemented.\n", __func__);

    return BadImplementation;
}


/*
 *
 */
_X_EXPORT Status
XvMCCompositeSubpicture(Display *display, XvMCSubpicture *subpicture,
			XvImage *image, short src_x, short src_y,
			unsigned short width, unsigned short height,
			short dst_x, short dst_y)
{
    fprintf(stderr, "%s is not implemented.\n", __func__);

    return BadImplementation;
}


/*
 *
 */
_X_EXPORT Status
XvMCDestroySubpicture(Display *display, XvMCSubpicture *subpicture)
{
    fprintf(stderr, "%s is not implemented.\n", __func__);

    return BadImplementation;
}


/*
 *
 */
_X_EXPORT Status
XvMCSetSubpicturePalette(Display *display, XvMCSubpicture *subpicture,
			 unsigned char *palette)
{
    fprintf(stderr, "%s is not implemented.\n", __func__);

    return BadImplementation;
}


/*
 *
 */
_X_EXPORT Status
XvMCBlendSubpicture(Display *display, XvMCSurface *surface,
		    XvMCSubpicture *subpicture, short sub_x, short sub_y,
		    unsigned short sub_width, unsigned short sub_h,
		    short surface_x, short surface_y,
		    unsigned short surface_width, unsigned short surface_h)
{
    fprintf(stderr, "%s is not implemented.\n", __func__);

    return BadImplementation;
}


/*
 *
 */
_X_EXPORT Status
XvMCBlendSubpicture2(Display *display, XvMCSurface *source, XvMCSurface *target,
		     XvMCSubpicture *subpicture, short sub_x, short sub_y,
		    unsigned short sub_width, unsigned short sub_h,
		    short surface_x, short surface_y,
		    unsigned short surface_width, unsigned short surface_h)
{
    fprintf(stderr, "%s is not implemented.\n", __func__);

    return BadImplementation;
}


/*
 *
 * These are empty ones and should not be used: Macroblock only.
 *
 */
/*
 *
 */
_X_EXPORT Status
XvMCCreateBlocks(Display *display, XvMCContext *context,
		 unsigned int num_blocks, XvMCBlockArray *blocks)
{
    fprintf(stderr, "%s is not implemented.\n", __func__);

    return BadImplementation;
}


/*
 *
 */
_X_EXPORT Status
XvMCDestroyBlocks(Display *display, XvMCBlockArray *blocks)
{
    fprintf(stderr, "%s is not implemented.\n", __func__);

    return Success;
}


/*
 *
 */
_X_EXPORT Status
XvMCCreateMacroBlocks(Display *display, XvMCContext *context,
		      unsigned int num_blocks, XvMCMacroBlockArray *blocks)
{
    fprintf(stderr, "%s is not implemented.\n", __func__);

    return BadImplementation;
}


/*
 *
 */
_X_EXPORT Status
XvMCDestroyMacroBlocks(Display *display, XvMCMacroBlockArray *blocks)
{
    fprintf(stderr, "%s is not implemented.\n", __func__);

    return Success;
}

/*
 *
 */
_X_EXPORT Status
XvMCRenderSurface(Display *display, XvMCContext *context,
		  unsigned int picture_structure, XvMCSurface *target,
		  XvMCSurface *past, XvMCSurface *future, unsigned int flags,
		  unsigned int num_macroblocks, unsigned int first_macroblock,
		  XvMCMacroBlockArray *macroblock_array, XvMCBlockArray *blocks)
{
    fprintf(stderr, "%s is not implemented.\n", __func__);

    return BadImplementation;
}
