/*
 * Copyright 2004-2009 Luc Verhaegen.
 * Copyright 2004-2005 The Unichrome Project  [unichrome.sf.net]
 * Copyright 1998-2003 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2003 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

/*************************************************************************
 *
 *  File:       via_driver.c
 *  Content:    XFree86 4.0 for VIA/S3G UniChrome
 *
 ************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if defined(XSERVER_LIBPCIACCESS) && defined(__linux__)
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#endif

#ifndef REINVENTED_RAC
#include "xf86Resources.h"
#include "xf86RAC.h"
#endif

#include "shadowfb.h"

#include "globals.h"

#ifdef HAVE_DPMSCONST_H
#include "X11/extensions/dpmsconst.h"
#else
#define DPMS_SERVER
#include "X11/extensions/dpms.h"
#endif

#include "via_driver.h"
#include "../git_version.h"

#ifndef _XF86_ANSIC_H
#include <string.h>
#endif

#ifdef HAVE_DRI
#define _XF86DRI_SERVER_
#include "dri.h"
#endif

#include "via_id.h"
#include "via_crtc.h"
#include "via_mode.h"

#include "xf86i2c.h"
#include "via_i2c.h"
#include "edid.h" /* for xf86MonPtr */
#include "via_output.h"

#include "via_2dregs.h"

#include "via_regs.h"

#include "vgaHW.h"
#include "mipointer.h"

#define USE_FB
#ifdef USE_FB
#include "fb.h"
#else
#include "cfb.h"
#include "cfb16.h"
#include "cfb32.h"
#endif

#include "xaa.h"

#ifdef REMOVED_LOADER_LISTS
#define xf86LoaderReqSymLists(p, n)
#endif

/*
 * prototypes
 */

static void VIAIdentify(int flags);
#ifdef XSERVER_LIBPCIACCESS
static Bool VIAPciProbe(DriverPtr drv, int entityNum, struct pci_device *dev,
			intptr_t matchData);
#else
static Bool VIAProbe(DriverPtr drv, int flags);
#endif
static Bool VIAPreInit(ScrnInfoPtr pScrn, int flags);
static Bool VIAEnterVT(int scrnIndex, int flags);
static void VIALeaveVT(int scrnIndex, int flags);
static Bool ViaMapFB(ScrnInfoPtr pScrn);
static void VIASave(ScrnInfoPtr pScrn);
static void VIARestore(ScrnInfoPtr pScrn);
static void VIAWriteMode(ScrnInfoPtr pScrn, DisplayModePtr mode);
static Bool VIACloseScreen(int scrnIndex, ScreenPtr pScreen);
static Bool VIAScreenInit(int scrnIndex, ScreenPtr pScreen, int argc,
                          char **argv);
static int VIAInternalScreenInit(ScrnInfoPtr pScrn, ScreenPtr pScreen);
static void VIAFreeScreen(int scrnIndex, int flags);
static Bool VIASwitchMode(int scrnIndex, DisplayModePtr mode, int flags);
static void VIAAdjustFrame(int scrnIndex, int y, int x, int flags);
static void VIADPMS(ScrnInfoPtr pScrn, int mode, int flags);
static const OptionInfoRec * VIAAvailableOptions(int chipid, int busid);

static Bool VIAMapMMIO(ScrnInfoPtr pScrn);
static void VIAUnmapMem(ScrnInfoPtr pScrn);
static void ViaFBEnable(ScrnInfoPtr pScrn);
static void ViaMMIOEnable(ScrnInfoPtr pScrn);

#ifndef _X_EXPORT
#define _X_EXPORT
#endif

#ifdef XSERVER_LIBPCIACCESS

# define PCI_ID_LIST_DECL struct pci_id_match VIADeviceMatch
# define VIA_DEVICE_MATCH(i, d) \
    { 0x1106, (d), PCI_MATCH_ANY, PCI_MATCH_ANY, 0, 0, (i) }
# define PCI_ID_LIST_END { 0, 0, (~0), (~0), 0, 0, 0 }

#else

# define PCI_ID_LIST_DECL PciChipsets VIAPciChipsets
resRange RES_SHARED_NONE[] = { _END };
# define VIA_DEVICE_MATCH(i, d) \
    { (i), (d), RES_SHARED_NONE }
# define PCI_ID_LIST_END { -1,	 -1,     RES_UNDEFINED }
#endif

/* Supported chipsets */

static SymTabRec VIAChipsets[] = {
    {VT3122,  "VT3122"},
    {VT7205,  "VT7205"},
    {VT3108,  "VT3108"},
    {VT3118,  "VT3118"},
    {VT3344,  "VT3344"},
    {VT3157,  "VT3157"},
    {VT3230,  "VT3230"},
    {VT3343,  "VT3343"},
    {VT3371,  "VT3371"},
    {VT1122,  "VT1122"},
    {VT5122,  "VT5122"},
    {-1,      NULL }
};

/* This table maps a PCI device ID to a chipset family identifier. */
static PCI_ID_LIST_DECL[] = {
    VIA_DEVICE_MATCH(VT3122,  PCI_CHIP_VT3122),
    VIA_DEVICE_MATCH(VT7205,  PCI_CHIP_VT7205),
    VIA_DEVICE_MATCH(VT3108,  PCI_CHIP_VT3108),
    VIA_DEVICE_MATCH(VT3118,  PCI_CHIP_VT3118),
    VIA_DEVICE_MATCH(VT3344,  PCI_CHIP_VT3344),
    VIA_DEVICE_MATCH(VT3157,  PCI_CHIP_VT3157),
    VIA_DEVICE_MATCH(VT3230,  PCI_CHIP_VT3230),
    VIA_DEVICE_MATCH(VT3343,  PCI_CHIP_VT3343),
    VIA_DEVICE_MATCH(VT3371,  PCI_CHIP_VT3371),
    VIA_DEVICE_MATCH(VT1122,  PCI_CHIP_VT1122),
    VIA_DEVICE_MATCH(VT5122,  PCI_CHIP_VT5122),
    PCI_ID_LIST_END
};


_X_EXPORT DriverRec UNICHROME =
{
    VIA_VERSION,
    "unichrome",
    VIAIdentify,
#ifdef XSERVER_LIBPCIACCESS
    NULL,
#else
    VIAProbe,
#endif
    VIAAvailableOptions,
    NULL,
    0,
    NULL,
#ifdef XSERVER_LIBPCIACCESS
    VIADeviceMatch,
    VIAPciProbe
#endif
};

typedef enum {
    OPTION_SIMULTANEOUS,
    OPTION_ACCELMETHOD,
    OPTION_SWCURSOR,
    OPTION_ROTATE,
#ifdef _VIA_VIDEO_H_
    OPTION_NOXV,
#endif
#ifdef HAVE_DRI
    OPTION_NODRI,
#endif
    OPTION_ACTIVEDEVICE,
    OPTION_CRTC2SCALE,
    OPTION_CRTC2HSCALE,
    OPTION_CRTC2VSCALE,
    OPTION_FBDIRECT,
    OPTION_DISABLEVQ,
    OPTION_DISABLEIRQ,
    OPTION_INSECUREDRI,
    OPTION_DISABLECOMMANDBUFFER,
    OPTION_DISABLECOMMANDBUFFERFOR2D,
    OPTION_USEDMACOPY
} VIAOpts;


static OptionInfoRec VIAOptions[] =
{
    {OPTION_SIMULTANEOUS, "Simultaneous", OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_ACCELMETHOD, "AccelMethod", OPTV_ANYSTR, {0}, FALSE},
    {OPTION_SWCURSOR,   "SWCursor",     OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_ROTATE,     "Rotate",       OPTV_ANYSTR,  {0}, FALSE},
#ifdef _VIA_VIDEO_H_
    {OPTION_NOXV,       "NoXv",         OPTV_BOOLEAN, {0}, FALSE},
#endif
#ifdef HAVE_DRI
    {OPTION_NODRI, "NoDRI", OPTV_BOOLEAN, {0}, FALSE},
#endif
    {OPTION_ACTIVEDEVICE, "ActiveDevice",       OPTV_ANYSTR,  {0}, FALSE},
    {OPTION_CRTC2SCALE, "CRTC2Scale", OPTV_ANYSTR, {0}, FALSE},
    {OPTION_CRTC2HSCALE, "CRTC2HScale", OPTV_ANYSTR, {0}, FALSE},
    {OPTION_CRTC2VSCALE, "CRTC2VScale", OPTV_ANYSTR, {0}, FALSE},
    {OPTION_FBDIRECT, "DirectFB", OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_DISABLEVQ,  "DisableVQ",    OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_DISABLEIRQ, "DisableIRQ", OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_DISABLECOMMANDBUFFER, "DisableCommandBuffer", OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_DISABLECOMMANDBUFFERFOR2D, "DisableCommandBufferFor2D", OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_USEDMACOPY, "UseDMACopy", OPTV_BOOLEAN, {0}, FALSE},
    {-1,                NULL,           OPTV_NONE,    {0}, FALSE}
};

#ifndef REMOVED_LOADER_LISTS
static const char *vgaHWSymbols[] = {
    "vgaHWGetHWRec",
    "vgaHWSetMmioFuncs",
    "vgaHWSave",
    "vgaHWProtect",
    "vgaHWRestore",
    "vgaHWSaveScreen",
    "vgaHWFreeHWRec",
    NULL
};


static const char *ramdacSymbols[] = {
    "xf86InitCursor",
    "xf86CreateCursorInfoRec",
    "xf86DestroyCursorInfoRec",
    NULL
};

static const char *ddcSymbols[] = {
    "xf86PrintEDID",
    "xf86DoEDID_DDC2",
    "xf86SetDDCproperties",
    NULL
};


static const char *i2cSymbols[] = {
    "xf86CreateI2CBusRec",
    "xf86I2CBusInit",
    "xf86CreateI2CDevRec",
    "xf86I2CDevInit",
    "xf86I2CWriteRead",
    "xf86I2CProbeAddress",
    "xf86DestroyI2CDevRec",
    "xf86I2CReadByte",
    "xf86I2CWriteByte",
    NULL
};

static const char *xaaSymbols[] = {
    "XAAGetCopyROP",
    "XAAGetCopyROP_PM",
    "XAAGetPatternROP",
    "XAACreateInfoRec",
    "XAADestroyInfoRec",
    "XAAInit",
    NULL
};

static const char *shadowSymbols[] = {
    "ShadowFBInit",
    NULL
};

#ifdef USE_FB
static const char *fbSymbols[] = {
    "fbScreenInit",
    "fbPictureInit",
    NULL
};
#else
static const char *cfbSymbols[] = {
    "cfbScreenInit",
    "cfb16ScreenInit",
    "cfb32ScreenInit",
    NULL
};
#endif

#endif /* REMOVED_LOADER_LISTS */

#ifdef XFree86LOADER

#ifndef REMOVED_LOADER_LISTS
#ifdef HAVE_DRI
static const char *drmSymbols[] = {
    "drmAddMap",
    "drmAgpAcquire",
    "drmAgpAlloc",
    "drmAgpBase",
    "drmAgpBind",
    "drmAgpEnable",
    "drmAgpFree",
    "drmAgpGetMode",
    "drmAgpRelease",
    "drmCtlInstHandler",
    "drmCtlUninstHandler",
    "drmCommandWrite",
    "drmCommandWriteRead",
    "drmGetInterruptFromBusID",
    "drmGetVersion",
    "drmMap",
    "drmUnmap",
    "drmAgpUnbind",
    "drmRmMap",
    NULL
};

static const char *driSymbols[] = {
    "DRICloseScreen",
    "DRICreateInfoRec",
    "DRIDestroyInfoRec",
    "DRIFinishScreenInit",
    "DRIGetSAREAPrivate",
    "DRILock",
    "DRIQueryVersion",
    "DRIScreenInit",
    "DRIUnlock",
    "GlxSetVisualConfigs",
    NULL
};
#endif /* HAVE_DRI */
#endif /* REMOVED_LOADER_LISTS */

static XF86ModuleVersionInfo UNICHROMEVersRec = {
    "unichrome",
    "http://unichrome.sf.net/",
    MODINFOSTRING1,
    MODINFOSTRING2,
#ifdef XORG_VERSION_CURRENT
    XORG_VERSION_CURRENT,
#else
    XF86_VERSION_CURRENT,
#endif
    VERSION_MAJOR, VERSION_MINOR, PATCHLEVEL,
    ABI_CLASS_VIDEODRV,
    ABI_VIDEODRV_VERSION,
    MOD_CLASS_VIDEODRV,
    {0, 0, 0, 0}
};

static pointer UNICHROMESetup(
#ifdef HaveModuleDescPtr /* XFree86 4.6.0 */
                        ModuleDescPtr module,
#else
                        pointer module,
#endif
                        pointer opts, int *errmaj, int *errmin)
{
    static Bool setupDone = FALSE;

    if (!setupDone) {
        setupDone = TRUE;
        xf86AddDriver(&UNICHROME, module, HaveDriverFuncs);
#ifndef REMOVED_LOADER_LISTS
        LoaderRefSymLists(vgaHWSymbols,
#ifdef USE_FB
                          fbSymbols,
#else
                          cfbSymbols,
#endif
                          ramdacSymbols,
                          xaaSymbols,
                          shadowSymbols,
                          i2cSymbols,
                          ddcSymbols,
#ifdef HAVE_DRI
			  drmSymbols,
			  driSymbols,
#endif
                          NULL);
#endif /* REMOVED_LOADER_LISTS */
        return (pointer) 1;
    }
    else {
        if (errmaj)
            *errmaj = LDR_ONCEONLY;

        return NULL;
    }
}

_X_EXPORT XF86ModuleData unichromeModuleData = {&UNICHROMEVersRec, UNICHROMESetup, NULL};

#endif /* XFree86LOADER */

static Bool
VIAGetRec(ScrnInfoPtr pScrn)
{
    VIAPtr pVia;

    VIAFUNC(pScrn);

    if (pScrn->driverPrivate)
        return TRUE;

    pScrn->driverPrivate = xnfcalloc(sizeof(struct VIARec), 1);
    pVia = VIAPTR(pScrn);
    pVia->scrnIndex = pScrn->scrnIndex;

    return TRUE;
}


static void
VIAFreeRec(ScrnInfoPtr pScrn)
{
    VIAPtr pVia;

    VIAFUNC(pScrn);

    if (!pScrn->driverPrivate)
        return;

    pVia = VIAPTR(pScrn);

    VIACursorDestroy(pVia);

    if (pVia->Crtc[0]) {
        if (pVia->Crtc[0]->Name)
            free(pVia->Crtc[0]->Name);
        free(pVia->Crtc[0]);
    }

    if (pVia->Crtc[1]) {
        if (pVia->Crtc[1]->Name)
            free(pVia->Crtc[1]->Name);
        free(pVia->Crtc[1]);
    }

    ViaOutputsDestroy(pScrn->scrnIndex);

    if (pVia->Scratch)
        free(pVia->Scratch);

    if (pVia->SavedReg.VGAFB)
        free(pVia->SavedReg.VGAFB);

    VIAUnmapMem(pScrn);

    free(pScrn->driverPrivate);
    pScrn->driverPrivate = NULL;
}


static const OptionInfoRec *
VIAAvailableOptions(int chipid, int busid)
{

    return VIAOptions;

}


/*
 * This implementation is far more flexible and informative than
 * xf86PrintChipSets. It allows for some semi-subversive VIA bashing too.
 */
static void
VIAIdentify(int flags)
{
    xf86Msg(X_INFO, "UNICHROME: driver for VIA UniChrome/Chrome integrated graphics.\n");
    xf86Msg(X_INFO, "    VT3122: CLE266 (CastleRock).\n");
    xf86Msg(X_INFO, "    VT7205: KM400, KM400A, KN400, P4M800 (Unichrome).\n");
    xf86Msg(X_INFO, "    VT3108: K8M800, K8N800, K8N800A (Unichrome Pro, aka Pro B).\n");
    xf86Msg(X_INFO, "    VT3118: CN400, PM800, PM880, PN800 (Unichrome Pro A).\n");
    xf86Msg(X_INFO, "    VT3344: VN800, P4M800Pro, CN700, P4M800CE (Unichrome Pro).\n");
    xf86Msg(X_INFO, "    VT3157: CX700, CX700M, VX700, CX700-M2 (Unichrome Pro).\n");
    xf86Msg(X_INFO, "    VT3230: K8M890 (Chrome9 IGP).\n");
    xf86Msg(X_INFO, "    VT3343: P4M890, VN890 (Unichrome Pro).\n");
    xf86Msg(X_INFO, "    VT3371: P4M900, VN896, CN896 (Chrome 9 HC IGP).\n");
    xf86Msg(X_INFO, "    VT1122: VX800, VX820 (Chrome 9 HC3 IGP).\n");
    xf86Msg(X_INFO, "    VT5122: VX855 (Chrome 9 HCM IGP) [UNRELEASED].\n");
    xf86Msg(X_INFO, "\n");
    xf86Msg(X_INFO, " Not all of the listed chips are actually supported at this time.\n");
    xf86Msg(X_INFO, "\n");

    xf86Msg(X_NOTICE, "VIA Technologies does not support or endorse this driver in any way.\n");
    xf86Msg(X_NOTICE, "For support, please refer to http://unichrome.sf.net/ or your X vendor.\n");

#if PATCHLEVEL == 0
    xf86Msg(X_INFO, "Using development version of the unichrome driver.\n");
# ifndef GIT_USED
    xf86Msg(X_INFO, "No git version information present.\n");
# else
#  ifdef GIT_UNCOMMITTED
    xf86Msg(X_INFO, "git SHA-ID: %s (on %s) with uncommitted changes.\n", GIT_SHAID, GIT_BRANCH);
#  else
    xf86Msg(X_INFO, "git SHA-ID: %s (on %s).\n", GIT_SHAID, GIT_BRANCH);
#  endif /* GIT_UNCOMMITTED */
# endif /* GIT_USED */
#endif /* PATCHLEVEL */
}

/*
 *
 */
static Bool
VIAChipSupported(int Chip, int Id)
{
    switch (Chip) {
    case VT3122:
    case VT7205:
    case VT3108:
	return TRUE;
    case VT3118:
    case VT3344:
    case VT3157:
    case VT3230:
    case VT3343:
    case VT3371:
    case VT1122:
	xf86Msg(X_WARNING, "VT%04X support is experimental.\n", Id);
	return TRUE;
    default:
	xf86Msg(X_ERROR, "VT%04X is not supported yet.\n", Id);
	return FALSE;
    }
}

/*
 * Wrap the probe between libpcicaccess and old pci infrastructure.
 */
#ifdef XSERVER_LIBPCIACCESS
/*
 *
 */
static Bool
VIAPciProbe(DriverPtr drv, int entityNum,
	    struct pci_device *dev, intptr_t matchData)
{
    ScrnInfoPtr pScrn;
    VIAPtr pVia;

    if (!VIAChipSupported(matchData, dev->device_id)) {
	ViaIDKnown(dev, matchData);
	return FALSE;
    }

    pScrn = xf86ConfigPciEntity(NULL, 0, entityNum, NULL,
				RES_SHARED_VGA, NULL, NULL, NULL, NULL);
    if (!pScrn)
	return FALSE;

    pScrn->driverVersion = VIA_VERSION;
    pScrn->driverName    = "unichrome";
    pScrn->name          = "UNICHROME";
    pScrn->Probe         = NULL;
    pScrn->PreInit       = VIAPreInit;
    pScrn->ScreenInit    = VIAScreenInit;
    pScrn->SwitchMode    = VIASwitchMode;
    pScrn->AdjustFrame   = VIAAdjustFrame;
    pScrn->EnterVT       = VIAEnterVT;
    pScrn->LeaveVT       = VIALeaveVT;
    pScrn->FreeScreen    = VIAFreeScreen;
    pScrn->ValidMode     = NULL; /* we do our own validation */

    if (!VIAGetRec(pScrn))
	return FALSE;

    pVia = VIAPTR(pScrn);

    pVia->PciInfo = dev;
    pVia->Chipset = matchData;

    return TRUE;
}
#else /* XSERVER_LIBPCIACCESS */

/*
 *
 */
static Bool
ViaScrnInfoSetup(DriverPtr drv, int entityIndex)
{
    ScrnInfoPtr pScrn = xf86AllocateScreen(drv, 0);

    pScrn = xf86ConfigPciEntity(pScrn, 0, entityIndex, VIAPciChipsets, 0,
				0, 0, 0, 0);
    if (!pScrn)
	return FALSE;

    pScrn->driverVersion = VIA_VERSION;
    pScrn->driverName = "unichrome";
    pScrn->name = "UNICHROME";
    pScrn->Probe = VIAProbe;
    pScrn->PreInit = VIAPreInit;
    pScrn->ScreenInit = VIAScreenInit;
    pScrn->SwitchMode = VIASwitchMode;
    pScrn->AdjustFrame = VIAAdjustFrame;
    pScrn->EnterVT = VIAEnterVT;
    pScrn->LeaveVT = VIALeaveVT;
    pScrn->FreeScreen = VIAFreeScreen;
    pScrn->ValidMode = NULL;

    return TRUE;
}

/*
 *
 */
static Bool
VIAProbe(DriverPtr drv, int flags)
{
    GDevPtr *devSections;
    int     *usedChips;
    int     numDevSections;
    int     numUsed;
    Bool    foundScreen = FALSE;
    int     i;

    /* sanity checks */
    if ((numDevSections = xf86MatchDevice("unichrome", &devSections)) <= 0)
        return FALSE;

    if (xf86GetPciVideoInfo() == NULL)
        return FALSE;

    numUsed = xf86MatchPciInstances("unichrome", PCI_VENDOR_VIA, VIAChipsets,
                                    VIAPciChipsets, devSections, numDevSections,
                                    drv, &usedChips);
    free(devSections);

    if (numUsed <= 0)
        return FALSE;

    if (flags & PROBE_DETECT) {
        foundScreen = TRUE;
    } else {
	for (i = 0; i < numUsed; i++) {
	    EntityInfoPtr pEnt = xf86GetEntityInfo(usedChips[i]);
	    pciVideoPtr pPci = xf86GetPciInfoForEntity(usedChips[i]);

	    if (!VIAChipSupported(pEnt->chipset, pPci->chipType)) {
		ViaIDKnown(pPci, pEnt->chipset);
		continue;
	    }

	    if (ViaScrnInfoSetup(drv, usedChips[i]))
		foundScreen = TRUE;
	}
    }

    free(usedChips);

    return foundScreen;
}
#endif

#ifdef XSERVER_LIBPCIACCESS
/*
 * People broke loads of things with libpciaccess, and never bothered
 * writing up proper replacements for it, even though, in some cases,
 * it was trivial to do so.
 */
#ifdef __linux__
/* go through sysfs to enable our device. */
static Bool
ViaPciDeviceEnableLinuxSysfs(VIAPtr pVia)
{
    char sysfsname[48];
    int fd, ret;

    VIAFUNC(pVia);

    snprintf(sysfsname, 47, "/sys/bus/pci/devices/%04x:%02x:%02x.%1u/enable",
	     pVia->PciInfo->domain, pVia->PciInfo->bus, pVia->PciInfo->dev,
	     pVia->PciInfo->func);

    fd = open(sysfsname, O_RDWR);
    if (fd == -1) {
	xf86DrvMsg(pVia->scrnIndex, X_ERROR, "%s: Failed to open %s: %s\n",
		   __func__, sysfsname, strerror(errno));
	return FALSE;
    }

    ret = write(fd, "1", 1);
    if (ret != 1) {
	xf86DrvMsg(pVia->scrnIndex, X_ERROR, "%s: Failed to write to %s: %s\n",
		       __func__, sysfsname, strerror(errno));
	close(fd);
	return FALSE;
    }

    close(fd);
    return TRUE;
}

/* go through sysfs to enable our device. */
static void
ViaPciDeviceDisableLinuxSysfs(VIAPtr pVia)
{
    char sysfsname[48];
    int fd, ret;

    VIAFUNC(pVia);

    snprintf(sysfsname, 47, "/sys/bus/pci/devices/%04x:%02x:%02x.%1u/enable",
	     pVia->PciInfo->domain, pVia->PciInfo->bus, pVia->PciInfo->dev,
	     pVia->PciInfo->func);

    fd = open(sysfsname, O_RDWR);
    if (fd == -1) {
	xf86DrvMsg(pVia->scrnIndex, X_ERROR, "%s: Failed to open %s: %s\n",
		   __func__, sysfsname, strerror(errno));
	return;
    }

    ret = write(fd, "1", 1);
    if (ret != 1) {
	xf86DrvMsg(pVia->scrnIndex, X_ERROR, "%s: Failed to write to %s: %s\n",
		       __func__, sysfsname, strerror(errno));
	close(fd);
	return;
    }

    close(fd);
}

#endif

static Bool
ViaPciDeviceEnable(VIAPtr pVia)
{
    VIAFUNC(pVia);

#ifdef __linux__
    if (ViaPciDeviceEnableLinuxSysfs(pVia))
	return TRUE;
#endif
    /* fallback, find out whether it is enabled. */
    {
	CARD16 Command;

	/* oh please, where are the nice libpci defines?
	   This is the crap you run into when you reinvent wheels */
	pci_device_cfg_read_u16(pVia->PciInfo, &Command, 4);
	if (Command & 0x02) /* we only need memory anyway. */
	    return TRUE;
	else
	    return FALSE;
    }
}

static void
ViaPciDeviceDisable(VIAPtr pVia)
{
    VIAFUNC(pVia);

#ifdef __linux__
    ViaPciDeviceDisableLinuxSysfs(pVia);
#endif
    return;
}

#else
static Bool
ViaPciDeviceEnable(VIAPtr pVia)
{
    /* life used to be so nice for the driver developer. */
    return TRUE;
}

static void
ViaPciDeviceDisable(VIAPtr pVia)
{
    return;
}
#endif

/*
 *
 */
static char *
ViaAccelMethodPrint(enum VIAAccelMethod AccelMethod)
{
    switch (AccelMethod) {
    case VIA_ACCEL_NONE:
	return "None";
    case VIA_ACCEL_SHADOWFB:
	return "ShadowFB";
    case VIA_ACCEL_XAA:
	return "XAA";
    case VIA_ACCEL_EXA:
	return "EXA";
    default:
	return "Unknown";
    }
}

/*
 *
 */
static void
ViaOptionAccelMethodHandle(VIAPtr pVia, OptionInfoRec *VIAOptions)
{
    char *AccelMethod;

    AccelMethod = xf86GetOptValString(VIAOptions, OPTION_ACCELMETHOD);
    if (!AccelMethod)
	pVia->AccelMethod = VIA_ACCEL_XAA;
    else {
	if (!strcasecmp(AccelMethod, "none"))
	    pVia->AccelMethod = VIA_ACCEL_NONE;
	else if (!strcasecmp(AccelMethod, "shadowfb"))
	    pVia->AccelMethod = VIA_ACCEL_SHADOWFB;
	else if (!strcasecmp(AccelMethod, "xaa"))
	    pVia->AccelMethod = VIA_ACCEL_XAA;
	else {
	    xf86DrvMsg(pVia->scrnIndex, X_WARNING, "Unknown AccelMethod \"%s\".\n",
		       AccelMethod);
	    pVia->AccelMethod = VIA_ACCEL_XAA;
	}
    }

    xf86DrvMsg(pVia->scrnIndex, X_CONFIG, "Attempting to use AccelMethod \"%s\"\n",
	       ViaAccelMethodPrint(pVia->AccelMethod));
}

/*
 * Concentrate all VIA scratch area stuff in one struct.
 * Or, keep all nastyness close together and far, far away.
 *
 * Known VIA Scratch Registers:
 *   SR32, SR33, SR34, SR39, SR3A
 *   CR3B, CR3C, CR3D, CR3E, CR3F,
 *   CR49, CR4A, CR4B, CR4C, CR4D, CR4E, CR4F.
 *
 */
static struct ViaScratch *
ViaScratchGet(VIAPtr pVia)
{
    struct ViaScratch *Scratch = xnfcalloc(sizeof(struct ViaScratch), 1);
    int tmp;

    VIAFUNC(pVia);

    if (!xf86IsScreenPrimary(pVia->scrnIndex))
        xf86DrvMsg(pVia->scrnIndex, X_INFO, "%s called when not the primary"
                   "VGA device. Some scratch areas might be uninitialised.\n",
                   __func__);

    /*
     * Get Memory size first. Should really use the RAM controller pci config.
     */
    if (pVia->Chipset == VT3122) {
        tmp = VGASRRead(pVia, 0x34);

        if (!tmp) {
            xf86DrvMsg(pVia->scrnIndex, X_WARNING, "%s: VideoRam Scratch area"
                       " uninitialised. Trying CR39.\n", __func__);
            tmp = VGASRRead(pVia, 0x39);
        }
    } else
        tmp = VGASRRead(pVia, 0x39);

    if ((tmp > 16) && (tmp <= 128))
        Scratch->VideoRam = (tmp + 1) << 9;
    else if ((tmp > 0) && (tmp < 31))
        Scratch->VideoRam = tmp << 12;
    else {
        xf86DrvMsg(pVia->scrnIndex, X_WARNING, "%s: VideoRam Scratch area"
                   " uninitialised.\n", __func__);
        Scratch->VideoRam = 16 << 10;
    }

    ViaDebug(pVia->scrnIndex, "%s: VideoRam: %d\n", __func__,
             Scratch->VideoRam);


    /*
     * Get the memory clock.
     */
    tmp = VGACRRead(pVia, 0x3D) >> 4;

    /* Actual values:
     *  0: SDR  66Mhz
     *  1: SDR 100Mhz
     *  2: SDR 133Mhz
     *  3: DDR 100Mhz (PC1600 or DDR200)
     *  4: DDR 133Mhz (PC2100 or DDR266)
     *  5: DDR 166Mhz (PC2700 or DDR333)
     *  6: DDR 200Mhz (PC3200 or DDR400)
     *  Guesses:
     *  7: DDR2 100Mhz (DDR2 400)
     *  8: DDR2 133Mhz (DDR2 533)
     *  9: DDR2 166Mhz (DDR2 667)
     *  A: DDR2 200Mhz (DDR2 800)
     *  B: DDR2 233Mhz (DDR2 1066)
     *  C: and above: Unknown.
     */
    /* SDR won't happen, assume badly set up register. */
    if ((tmp < 0x03) || (tmp > 0x0A)) {
        xf86DrvMsg(pVia->scrnIndex, X_WARNING, "%s: Unknown Memory Clock (%d)."
                   " Defaulting to DDR200\n", __func__, tmp);
        Scratch->MemType = VIA_MEM_DDR_200;
    } else
        Scratch->MemType = tmp - 3;
    ViaDebug(pVia->scrnIndex, "%s: Memory Type: %d\n", __func__, Scratch->MemType);

    /*
     * Get output device status.
     */
    Scratch->ActiveDevice = VGACRRead(pVia, 0x3E) >> 4;
    ViaDebug(pVia->scrnIndex, "%s: ActiveDevice: %d\n", __func__,
             Scratch->ActiveDevice);

    /*
     * Get Panel size.
     */
    Scratch->PanelSize = VGACRRead(pVia, 0x3F) >> 4;
    ViaDebug(pVia->scrnIndex, "%s: PanelSize: %d\n",
             __func__, Scratch->PanelSize);
    if (!Scratch->PanelSize) {
        if (VGACRRead(pVia, 0x6A) & 0x08) /* LCD enabled? */
            xf86DrvMsg(pVia->scrnIndex, X_WARNING, "%s: It is possible that"
                       " PanelSize wasn't initialised properly\n", __func__);
    }

    /*
     * Get TV Standard.
     */
    if (VGACRRead(pVia, 0x3B) & 0x02)
        Scratch->TVStandard = TVSTANDARD_PAL;
    else
        Scratch->TVStandard = TVSTANDARD_NTSC;
    ViaDebug(pVia->scrnIndex, "%s: TVStandard: %d\n", __func__,
             Scratch->TVStandard);

    return Scratch;
}

/*
 * Handle the options for the type of scaling happening on CRTC2.
 * Fugly.
 */
static void
VIACrtc2OptionScaling(VIAPtr pVia)
{
    char *s;

    pVia->Crtc2HScaleInterpolate = TRUE;
    pVia->Crtc2VScaleInterpolate = TRUE;

    /* General scaling overrules all */
    if ((s = xf86GetOptValString(VIAOptions, OPTION_CRTC2SCALE))) {
	if (!xf86NameCmp(s, "duplicate") ||
	    !xf86NameCmp(s, "duplication")) {
	    pVia->Crtc2HScaleInterpolate = FALSE;
	    pVia->Crtc2VScaleInterpolate = FALSE;
	    xf86DrvMsg(pVia->scrnIndex, X_CONFIG,
                       "Crtc2: using hardware duplication when scaling.\n");
	} else if (!xf86NameCmp(s, "interpolate") ||
		   !xf86NameCmp(s, "interpolation"))
	    xf86DrvMsg(pVia->scrnIndex, X_CONFIG,
		       "Crtc2: using hardware interpolation when scaling.\n");
	else {
	    xf86DrvMsg(pVia->scrnIndex, X_WARNING,
		       "Crtc2: unable to parse Option CRTC2Scale.\n");
	    xf86DrvMsg(pVia->scrnIndex, X_CONFIG,
		       "Crtc2: defaulting to hardware interpolation when scaling.\n");
	}
    } else {
	/* Horizontal scaling */
	if ((s = xf86GetOptValString(VIAOptions, OPTION_CRTC2HSCALE))) {
	    if (!xf86NameCmp(s, "duplicate") ||
		!xf86NameCmp(s, "duplication")) {
		pVia->Crtc2HScaleInterpolate = FALSE;
		xf86DrvMsg(pVia->scrnIndex, X_CONFIG,
			   "Crtc2: using hardware duplication when scaling horizontally.\n");
	    } else if (!xf86NameCmp(s, "interpolate") ||
		       !xf86NameCmp(s, "interpolation"))
		xf86DrvMsg(pVia->scrnIndex, X_CONFIG,
			   "Crtc2: using hardware interpolation when scaling horizontally.\n");
	    else {
		xf86DrvMsg(pVia->scrnIndex, X_WARNING,
			   "Crtc2: unable to parse Option CRTC2HScale.\n");
		xf86DrvMsg(pVia->scrnIndex, X_CONFIG,
			   "Crtc2: defaulting to hardware interpolation when scaling horizontally.\n");
	    }
	} else
	    xf86DrvMsg(pVia->scrnIndex, X_CONFIG,
		       "Crtc2: using hardware interpolation when scaling horizontally.\n");

	/* Vertical scaling */
	if ((s = xf86GetOptValString(VIAOptions, OPTION_CRTC2VSCALE))) {
	    if (!xf86NameCmp(s, "duplicate") ||
		!xf86NameCmp(s, "duplication")) {
		pVia->Crtc2VScaleInterpolate = FALSE;
		xf86DrvMsg(pVia->scrnIndex, X_CONFIG,
			   "Crtc2: using hardware duplication when scaling vertically.\n");
	    } else if (!xf86NameCmp(s, "interpolate") ||
		       !xf86NameCmp(s, "interpolation"))
		xf86DrvMsg(pVia->scrnIndex, X_CONFIG,
			   "Crtc2: using hardware interpolation when scaling vertically.\n");
	    else {
		xf86DrvMsg(pVia->scrnIndex, X_WARNING,
			   "Crtc2: unable to parse Option CRTC2VScale.\n");
		xf86DrvMsg(pVia->scrnIndex, X_CONFIG,
			   "Crtc2: defaulting to hardware interpolation when scaling vertically.\n");
	    }
	} else
	    xf86DrvMsg(pVia->scrnIndex, X_CONFIG,
		       "Crtc2: using hardware interpolation when scaling vertically.\n");
    }
}

/*
 * Work around hardware limitations for CRTC2.
 */
static void
PreInitCrtc2Quirks(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);

    /* Bandwidth issues are currently mitigated by having only
       a single CRTC active at a time. */

    /* Cursor is handled in via_cursor.c */

#ifdef _VIA_VIDEO_H_
    /* Video3 doesn't display on the secondary CRTC on VT3122 */
    if (pVia->UseXv) {
        if (pVia->Chipset == VT3122) {
            xf86DrvMsg(pScrn->scrnIndex, X_WARNING, "Hardware Video Overlay (3)"
                       " is not supported on the Secondary CRTC.\n");
            xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
                       "X Video support disabled.\n");
            pVia->UseXv = FALSE;
        } else
            pVia->SwovCrtc2 = TRUE;
    }
#endif
}

/*
 *
 */
static Bool
VIAPreInit(ScrnInfoPtr pScrn, int flags)
{
    VIAPtr          pVia;
    char            *s = NULL;
#ifndef USE_FB
    char            *mod = NULL;
    const char      *reqSym = NULL;
#endif
    DisplayModePtr  Modes;

    VIAFUNC(pScrn);

    if (flags & PROBE_DETECT) {
	xf86DrvMsg(pScrn->scrnIndex, X_INFO,
		   "Skipping DDC probing in xserver configure stage.\n"
		   "    libv does not believe in preconfigured monitors; "
		   "They should only be used\n    as a last resort.\n");
        return FALSE;
    }

    if (pScrn->numEntities > 1)
        return FALSE;

    if (!xf86LoadSubModule(pScrn, "vgahw"))
        return FALSE;

    xf86LoaderReqSymLists(vgaHWSymbols, NULL);
    if (!vgaHWGetHWRec(pScrn))
        return FALSE;

    if (!VIAGetRec(pScrn))
        return FALSE;
    pVia = VIAPTR(pScrn);

#ifdef XSERVER_LIBPCIACCESS
    pVia->ChipId = pVia->PciInfo->vendor_id;
    pScrn->chipset = (char *)xf86TokenToString(VIAChipsets, pVia->Chipset);
#else
    {
	EntityInfoPtr pEnt = xf86GetEntityInfo(pScrn->entityList[0]);
	if (pEnt->resources) {
	    free(pEnt);
	    VIAFreeRec(pScrn);
	    return FALSE;
	}

	pVia->PciInfo = xf86GetPciInfoForEntity(pEnt->index);
	xf86RegisterResources(pEnt->index, NULL, ResNone);
	pVia->ChipId = pVia->PciInfo->chipType;
	pVia->Chipset = pEnt->chipset;
	pScrn->chipset = (char *)xf86TokenToString(VIAChipsets, pVia->Chipset);
	free(pEnt);
    }
#endif

    xf86DrvMsg(pScrn->scrnIndex, X_PROBED, "Found %s UniChrome/Chrome IGP\n",
               pScrn->chipset);
    ViaHostIdentify(pScrn);
    ViaCheckCardId(pScrn);

#ifndef XSERVER_LIBPCIACCESS
    pVia->PciTag =
        pciTag(pVia->PciInfo->bus, pVia->PciInfo->device, pVia->PciInfo->func);
#endif

    if (!ViaPciDeviceEnable(pVia)) {
	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
		   "%s: Failed to enable pci device.\n", __func__);
	VIAFreeRec(pScrn);
	return FALSE;
    }

    if (!VIAMapMMIO(pScrn)) {
	ViaPciDeviceDisable(pVia);
	VIAFreeRec(pScrn);
        return FALSE;
    }
    ViaMMIOEnable(pScrn);

    /*
     * We support depths of 8, 16 and 24.
     * We support bpp of 8, 16, and 32.
     */

    if (!xf86SetDepthBpp(pScrn, 0, 0, 0, Support32bppFb)) {
	ViaPciDeviceDisable(pVia);
	VIAFreeRec(pScrn);
        return FALSE;
    }
    else {
        switch (pScrn->depth) {
        case 8:
        case 16:
        case 24:
        case 32:
            /* OK */
            break;
        default:
            xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                       "Given depth (%d) is not supported by this driver\n",
                       pScrn->depth);
	    ViaPciDeviceDisable(pVia);
	    VIAFreeRec(pScrn);
            return FALSE;
        }
    }

    xf86PrintDepthBpp(pScrn);

    if (pScrn->depth == 32)
        pScrn->depth = 24;

    if (pScrn->depth > 8) {
        rgb zeros = {0, 0, 0};

        if (!xf86SetWeight(pScrn, zeros, zeros)) {
	    ViaPciDeviceDisable(pVia);
	    VIAFreeRec(pScrn);
            return FALSE;
	} else {
            /* TODO check weight returned is supported */
            ;
        }
    }

    if (!xf86SetDefaultVisual(pScrn, -1)) {
        return FALSE;
    } else {
        /* We don't currently support DirectColor at > 8bpp */
        if (pScrn->depth > 8 && pScrn->defaultVisual != TrueColor) {
            xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Given default visual"
                       " (%s) is not supported at depth %d\n",
                       xf86GetVisualName(pScrn->defaultVisual), pScrn->depth);
	    ViaPciDeviceDisable(pVia);
	    VIAFreeRec(pScrn);
            return FALSE;
        }
    }

    /* We use a programmable clock */
    pScrn->progClock = TRUE;

    /* xf86CollectOptions depends on pScrn->monitor instead of
       pScrn->confScreen->monitor */
    pScrn->monitor = pScrn->confScreen->monitor;
    xf86CollectOptions(pScrn, NULL);
    pScrn->monitor = NULL; /* we might end up using a different MonRec anyway */

    /* Set the bits per RGB for 8bpp mode */
    if (pScrn->depth == 8)
        pScrn->rgbBits = 6;

    xf86ProcessOptions(pScrn->scrnIndex, pScrn->options, VIAOptions);

    ViaOptionAccelMethodHandle(pVia, VIAOptions);

    if (xf86ReturnOptValBool(VIAOptions, OPTION_SWCURSOR, FALSE)) {
        pVia->hwcursor = FALSE;
	xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Disabling hardware cursor.\n");
    } else
	pVia->hwcursor = TRUE;

    if ((s = xf86GetOptValString(VIAOptions, OPTION_ROTATE))) {
        if (!xf86NameCmp(s, "CW")) {
	    pVia->rotate = 1;
	    xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,
		       "Rotation selected: Clockwise.\n");
	} else if (!xf86NameCmp(s, "CCW")) {
	    pVia->rotate = -1;
	    xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,
		       "Rotation selected: Counter Clockwise.\n");
	} else {
	    xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
		       "Unknown Rotate option: \"%s\"\n", s);
	    pVia->rotate = 0;
	}
    } else
	pVia->rotate = 0;

    if (pVia->rotate) {
	switch (pVia->AccelMethod) {
	case VIA_ACCEL_XAA:
	case VIA_ACCEL_EXA:
	    xf86DrvMsg(pScrn->scrnIndex, X_INFO,
		       "Rotation selected: Disabling \"%s\" Acceleration.\n",
		       ViaAccelMethodPrint(pVia->AccelMethod));
	    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Using ShadowFB instead.\n");
	    pVia->AccelMethod = VIA_ACCEL_SHADOWFB;
	    break;
	case VIA_ACCEL_SHADOWFB:
	    break;
	case VIA_ACCEL_NONE:
	    xf86DrvMsg(pScrn->scrnIndex, X_INFO,
		       "\"%s\" Acceleration selected: Disabling rotation.\n",
		       ViaAccelMethodPrint(pVia->AccelMethod));

	    pVia->rotate = 0;
	    break;
	default:
	    xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
		       "Rotation selected: Unhandled AccelMethod %d.\n",
		       pVia->AccelMethod);
	    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Using ShadowFB instead.\n");
	    pVia->AccelMethod = VIA_ACCEL_SHADOWFB;
	    break;
	}
    }

    if (pVia->rotate && pVia->hwcursor) {
	xf86DrvMsg(pScrn->scrnIndex, X_INFO,
		   "Rotation selected: Disabling Hardware Cursor\n");
	pVia->hwcursor = FALSE;
    }

    if (pVia->hwcursor) {
        if (xf86LoadSubModule(pScrn, "ramdac"))
	    xf86LoaderReqSymLists(ramdacSymbols, NULL);
	else {
	    xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Failed to load ramdac module.\n");
	    xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Falling back to Software cursor.\n");
	    pVia->hwcursor = FALSE;
	}
    }

    if (pVia->AccelMethod == VIA_ACCEL_XAA) {
        if (xf86LoadSubModule(pScrn, "xaa"))
	    xf86LoaderReqSymLists(xaaSymbols, NULL);
	else {
	    xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Failed to load xaa module.\n");
	    xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Falling back to ShadowFB.\n");
	    pVia->AccelMethod = VIA_ACCEL_SHADOWFB;
	}
    }

    if (pVia->AccelMethod == VIA_ACCEL_SHADOWFB) {
        if (xf86LoadSubModule(pScrn, "shadowfb"))
	    xf86LoaderReqSymLists(shadowSymbols, NULL);
	else {
	    xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Failed to load shadowfb module.\n");
	    xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Disabling all forms of acceleration.\n");
	    pVia->AccelMethod = VIA_ACCEL_NONE;
	}
    }

#ifdef _VIA_VIDEO_H_
    if (xf86ReturnOptValBool(VIAOptions, OPTION_NOXV, FALSE)) {
        pVia->UseXv = FALSE;
        xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Disabling Xv from config.\n");
    } else
        pVia->UseXv = TRUE;
#endif

#ifdef HAVE_DRI
    if (xf86ReturnOptValBool(VIAOptions, OPTION_NODRI, FALSE)) {
        pVia->UseDRI = FALSE;
        xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Disabling DRI from config.\n");
    } else
        pVia->UseDRI = TRUE;
#endif

    if (xf86ReturnOptValBool(VIAOptions, OPTION_DISABLEVQ, FALSE)) {
	pVia->VQEnable = FALSE;
	xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Disabling Virtual Queue.\n");
    } else
	pVia->VQEnable = TRUE;

    if (xf86ReturnOptValBool(VIAOptions, OPTION_DISABLEIRQ, FALSE)) {
	pVia->DRIIrqEnable = FALSE;
	xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,
		   "Option: DisableIRQ - DRI IRQ Disabled\n");
    } else
        pVia->DRIIrqEnable = TRUE;

    if (xf86ReturnOptValBool(VIAOptions, OPTION_DISABLECOMMANDBUFFER, FALSE)) {
	pVia->UseCommandBuffer = FALSE;
	xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,
		   "Option: Disabling Command Buffer.\n");
    } else
	pVia->UseCommandBuffer = TRUE;

    if (pVia->UseCommandBuffer) {
	if (xf86ReturnOptValBool(VIAOptions, OPTION_DISABLECOMMANDBUFFERFOR2D, FALSE)) {
	    pVia->UseCommandBufferFor2d = FALSE;
	    xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,
		       "Option: Not using the Command Buffer for 2D acceleration.\n");
	} else
	    pVia->UseCommandBufferFor2d = TRUE;
    } else
	pVia->UseCommandBufferFor2d = FALSE;

#ifdef HAVE_DRI
    if (xf86ReturnOptValBool(VIAOptions, OPTION_USEDMACOPY, TRUE))
	pVia->UseDMACopy = TRUE; /* properly checked from DRIScreenInit */
    else
#endif
	pVia->UseDMACopy = FALSE;

    /* Simultaneous is a special mode where both CRTCs are tightly interlocked.
       This allows all overlays to function on both heads, but breaks scaling. */
    if (xf86ReturnOptValBool(VIAOptions, OPTION_SIMULTANEOUS, FALSE))
        pVia->Simultaneous = TRUE;
    else
        pVia->Simultaneous = FALSE;

    /* ActiveDevice Option for device selection */
    pVia->ActiveDevice = 0x00;
    if ((s = xf86GetOptValString(VIAOptions, OPTION_ACTIVEDEVICE))) {
	if (strstr(s, "CRT"))
	    pVia->ActiveDevice |= VIA_DEVICE_CRT;
	if (strstr(s, "DFP")) /* just treat this the same as LCD */
	    pVia->ActiveDevice |= VIA_DEVICE_LCD;
	if (strstr(s, "TV"))
	    pVia->ActiveDevice |= VIA_DEVICE_TV;
    }

    /* Do we set/force the Direct FB option? */
    pVia->FBDirectOption = xf86GetOptValBool(VIAOptions, OPTION_FBDIRECT,
					     &pVia->FBDirect);
    pVia->Scratch = ViaScratchGet(pVia);

    if (!ViaFBInit(pScrn)) {
	ViaPciDeviceDisable(pVia);
	VIAFreeRec(pScrn);
        return FALSE;
    }
    pVia->FBFreeStart = 0;
    pVia->FBFreeSize = pScrn->videoRam << 10;
    ViaDebug(pScrn->scrnIndex, "%s: Initially available framebuffer: 0x%08X "
	     "(0x%08X)\n", __func__, pVia->FBFreeStart, pVia->FBFreeSize);

    if (!xf86LoadSubModule(pScrn, "i2c")) {
	ViaPciDeviceDisable(pVia);
	VIAFreeRec(pScrn);
        return FALSE;
    } else {
        xf86LoaderReqSymLists(i2cSymbols,NULL);
        ViaI2CInit(pScrn);
    }

    if (!xf86LoadSubModule(pScrn, "ddc")) {
	ViaPciDeviceDisable(pVia);
	VIAFreeRec(pScrn);
	return FALSE;
    } else
	xf86LoaderReqSymLists(ddcSymbols, NULL);

    if (pVia->hwcursor && !VIACursorFBGrab(pScrn))
	pVia->hwcursor = FALSE;

    if (!ViaCrtcInit(pScrn)) {
	ViaPciDeviceDisable(pVia);
	VIAFreeRec(pScrn);
	return FALSE;
    }

    ViaOutputsDetect(pScrn->scrnIndex);
    if (!ViaLayOutCrtcsAndOutputs(pScrn->scrnIndex)) {
	xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "No outputs possible.\n");
	ViaPciDeviceDisable(pVia);
	VIAFreeRec(pScrn);
	return FALSE;
    }

    if (pVia->Crtc[1]->Active && !pVia->Crtc[0]->Active)
        PreInitCrtc2Quirks(pScrn);

    /* Only handle the scaling options when CRTC2 is actually active. */
    if (pVia->Crtc[1]->Active)
	VIACrtc2OptionScaling(pVia);

    ViaOutputTimingSetFromConfig(pScrn);

    pScrn->xInc = 8;

    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Clock range: %6.2f to %6.2f MHz\n",
               20.000, 230.000);

    pVia->XModes = pScrn->confScreen->monitor->Modes;

    if (pScrn->display->virtualX && pScrn->display->virtualY)
        ViaVirtualGetFromConfig(pScrn);

    ViaModeFixedInit(pScrn); /* For fixed timing modes */

    Modes = ViaModesPoolCreate(pScrn, FALSE);
    if (!Modes) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "No valid modes found\n");
	ViaPciDeviceDisable(pVia);
	VIAFreeRec(pScrn);
        return FALSE;
    }
    ViaModesAttach(pScrn, Modes);

    if (!pScrn->virtualX || !pScrn->virtualY)
        ViaVirtualGetFromModes(pScrn);

    pVia->FBScanoutStart = pVia->FBFreeStart;
    pVia->FBScanoutSize = (pScrn->displayWidth * pScrn->bitsPerPixel >> 3)
	* pScrn->virtualY;
    pVia->FBFreeStart += pVia->FBScanoutSize;
    pVia->FBFreeSize -= pVia->FBScanoutSize;

    ViaDebug(pScrn->scrnIndex, "%s: Scanout memory at 0x%08X (0x%08X)\n",
	     __func__, pVia->FBScanoutStart, pVia->FBScanoutSize);
    xf86DrvMsg(pScrn->scrnIndex, X_INFO,
               "Using %dx%d Framebuffer with %d pitch\n", pScrn->virtualX,
               pScrn->virtualY, pScrn->displayWidth);

    xf86PrintModes(pScrn);

#ifdef USE_FB
    if (xf86LoadSubModule(pScrn, "fb") == NULL) {
	ViaPciDeviceDisable(pVia);
	VIAFreeRec(pScrn);
        return FALSE;
    }

    xf86LoaderReqSymLists(fbSymbols, NULL);

#else
    /* load bpp-specific modules */
    switch (pScrn->bitsPerPixel) {
        case 8:
            mod = "cfb";
            reqSym = "cfbScreenInit";
            break;
        case 16:
            mod = "cfb16";
            reqSym = "cfb16ScreenInit";
            break;
        case 32:
            mod = "cfb32";
            reqSym = "cfb32ScreenInit";
            break;
    }

    if (mod && xf86LoadSubModule(pScrn, mod) == NULL) {
	ViaPciDeviceDisable(pVia);
	VIAFreeRec(pScrn);
        return FALSE;
    }

    xf86LoaderReqSymbols(reqSym, NULL);
#endif

    /* set dpi and windowsize ourselves.
     * need to dig this out properly: what gets influenced here?
     */
    pScrn->widthmm = pScrn->virtualX / 254;
    pScrn->heightmm = pScrn->virtualY / 254;
    pScrn->xDpi = 100;
    pScrn->yDpi = 100;

    /* set gamma to the default ourselves */
    pScrn->gamma.red = 1.0;
    pScrn->gamma.green = 1.0;
    pScrn->gamma.blue = 1.0;

    /* Setting SubpixelOrder to None, can't be bypassed. Thanks so very much. */
    pScrn->monitor = xnfcalloc(1, sizeof(MonRec));
    /* Where do we free this crap then? */

    if (pVia->AccelMethod == VIA_ACCEL_XAA)
	VIAAccelFBGrab(pScrn);
    if (pVia->VQEnable && !VIAVQFBGrab(pScrn))
	pVia->VQEnable = FALSE;

    ViaDebug(pScrn->scrnIndex, "%s: Remaining FB at 0x%08X (0x%08X)\n",
	     __func__, pVia->FBFreeStart, pVia->FBFreeSize);

    VIAUnmapMem(pScrn);
    ViaPciDeviceDisable(pVia);

    return TRUE;
}


static Bool
VIAEnterVT(int scrnIndex, int flags)
{
    ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
    VIAPtr pVia = VIAPTR(pScrn);

    /* FIXME: Rebind AGP memory here */
    VIAFUNC(pScrn);

    ViaMMIOEnable(pScrn);

    VIASave(pScrn);

    ViaFBEnable(pScrn);

    ViaCrtcModeSetInitial(pScrn);

    if (pVia->Crtc[0]->Active)
        ViaOutputsPower(pVia->Crtc[0], FALSE);
    if (pVia->Crtc[1]->Active)
        ViaOutputsPower(pVia->Crtc[1], FALSE);

    VIAWriteMode(pScrn, pScrn->currentMode);

    /* Patch for APM suspend resume, HWCursor has garbage */
    VIACursorRestore(pVia);

    /* Take 3D engine out of reset -- also needed for VQ.
     * So, hopefully, blindly poking it won't hurt */
    VGASRMask(pScrn, 0x1A, 0x00, 0x40);

#ifdef HAVE_DRI
    if (pVia->directRenderingEnabled) {
	if (pVia->UseCommandBuffer)
	    VIADRMCommandBufferStart(pVia);
	DRIUnlock(screenInfo.screens[scrnIndex]);
    }
#endif

    /* How does this interact with DRI? */
    if (pVia->VQEnable)
        ViaVQEnable(pScrn);

    if (pVia->Crtc[0]->Active)
        ViaOutputsPower(pVia->Crtc[0], TRUE);
    if (pVia->Crtc[1]->Active)
        ViaOutputsPower(pVia->Crtc[1], TRUE);

    return TRUE;
}


static void VIALeaveVT(int scrnIndex, int flags)
{
    ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
    VIAPtr pVia = VIAPTR(pScrn);

    VIAFUNC(pScrn);

#ifdef HAVE_DRI
    if (pVia->directRenderingEnabled)
	DRILock(pScrn->pScreen, 0);
#endif

    VIAAccelSync(pScrn);

    /* Stop 3D engine from hanging on VT switch: software reset */
    VGASRMask(pScrn, 0x1A, 0x40, 0x40);

#ifdef HAVE_DRI
    if (pVia->directRenderingEnabled && pVia->CommandBufferActive)
	VIADRMCommandBufferStop(pVia);
#endif

    if (pVia->VQEnable)
	ViaVQDisable(pScrn);

    VIACursorStore(pVia);

    VIARestore(pScrn);
}

/*
 *
 */
static void
VIASave(ScrnInfoPtr pScrn)
{
    vgaHWPtr  hwp = VGAHWPTR(pScrn);
    VIAPtr  pVia = VIAPTR(pScrn);
    VIARegPtr  Regs = &pVia->SavedReg;
    int i;

    VIAFUNC(pScrn);

    vgaHWProtect(pScrn, TRUE);

    vgaHWSave(pScrn, &hwp->SavedReg, VGA_SR_MODE);

    /* Unlock Extended Regs */
    VGASRWrite(pScrn, 0x10, 0x01);
    Regs->SR14 = VGASRRead(pScrn, 0x14);
    Regs->SR15 = VGASRRead(pScrn, 0x15);
    Regs->SR16 = VGASRRead(pScrn, 0x16);
    Regs->SR17 = VGASRRead(pScrn, 0x17);
    Regs->SR18 = VGASRRead(pScrn, 0x18);
    Regs->SR19 = VGASRRead(pScrn, 0x19);
    Regs->SR1A = VGASRRead(pScrn, 0x1A);
    Regs->SR1B = VGASRRead(pScrn, 0x1B);
    Regs->SR1C = VGASRRead(pScrn, 0x1C);
    Regs->SR1D = VGASRRead(pScrn, 0x1D);
    Regs->SR1E = VGASRRead(pScrn, 0x1E);
    Regs->SR1F = VGASRRead(pScrn, 0x1F);

    Regs->SR22 = VGASRRead(pScrn, 0x22);
    Regs->SR23 = VGASRRead(pScrn, 0x23);
    Regs->SR24 = VGASRRead(pScrn, 0x24);
    Regs->SR25 = VGASRRead(pScrn, 0x25);
    Regs->SR26 = VGASRRead(pScrn, 0x26);
    Regs->SR27 = VGASRRead(pScrn, 0x27);
    Regs->SR28 = VGASRRead(pScrn, 0x28);
    Regs->SR29 = VGASRRead(pScrn, 0x29);
    Regs->SR2A = VGASRRead(pScrn, 0x2A);
    Regs->SR2B = VGASRRead(pScrn, 0x2B);

    Regs->SR2E = VGASRRead(pScrn, 0x2E);

    Regs->SR44 = VGASRRead(pScrn, 0x44);
    Regs->SR45 = VGASRRead(pScrn, 0x45);
    Regs->SR46 = VGASRRead(pScrn, 0x46);
    Regs->SR47 = VGASRRead(pScrn, 0x47);
    Regs->SR4A = VGASRRead(pScrn, 0x4A);
    Regs->SR4B = VGASRRead(pScrn, 0x4B);
    Regs->SR4C = VGASRRead(pScrn, 0x4C);

    Regs->CR13 = VGACRRead(pScrn, 0x13);

    Regs->CR32 = VGACRRead(pScrn, 0x32);
    Regs->CR33 = VGACRRead(pScrn, 0x33);
    Regs->CR34 = VGACRRead(pScrn, 0x34);
    Regs->CR35 = VGACRRead(pScrn, 0x35);
    Regs->CR36 = VGACRRead(pScrn, 0x36);

    ViaOutputsSave(pScrn->scrnIndex);

    /* Save LCD control regs */
    for (i = 0; i < 68; i++)
	Regs->CRTCRegs[i] = VGACRRead(pScrn, i + 0x50);

    switch (pVia->Chipset) {
    case VT3108:
    case VT3344:
	Regs->CR97 = VGACRRead(pScrn, 0x97);
	Regs->CR99 = VGACRRead(pScrn, 0x99);
	break;
    default:
	break;
    }

    vgaHWProtect(pScrn, FALSE);

    /* If we're using secondary for textmode, then, upon restoration,
     * the standard VGA clock set through Misc will ruin the mode.
     */
    if (((Regs->CRTCRegs[0x1A] & 0xC0) == 0x40) && /* CRTC2 active */
	(Regs->CRTCRegs[0x1B] & 0x08) && /* Simultaneous */
	((hwp->SavedReg.MiscOutReg & 0x0C) != 0x0C)) { /* VGA clock set */
	xf86DrvMsg(pScrn->scrnIndex, X_WARNING, "Console uses both CRTCs"
		   " while depending on VGA clock selection.\n");
	xf86DrvMsg(pScrn->scrnIndex, X_WARNING, "Upon restoration, console "
		   "will not be useful without a workaround.\n");
	xf86DrvMsg(pScrn->scrnIndex, X_WARNING, "This workaround means that"
		   " future VGA mode changes might break.\n");
	hwp->SavedReg.MiscOutReg |= 0x0C;
    }

    /* Grab VGA framebuffer, the easy way, plenty fast too */
    if (ViaMapFB(pScrn)) {
	if (!Regs->VGAFB)
	    Regs->VGAFB = xnfcalloc(1, 256*1024);

	memcpy(Regs->VGAFB, pVia->FBBase, 256*1024);
    }
}

static void
VIARestore(ScrnInfoPtr pScrn)
{
    VIAPtr  pVia = VIAPTR(pScrn);
    VIARegPtr  Regs = &pVia->SavedReg;
    int  i;
    CARD8  tmp;

    VIAFUNC(pScrn);

    vgaHWProtect(pScrn, TRUE);
    /* Unlock Extended Regs */
    VGASRWrite(pScrn, 0x10, 0x01);
    VGACRMask(pScrn, 0x47, 0x00, 0x01);

    VGACRWrite(pScrn, 0x6A, 0x00);
    VGACRWrite(pScrn, 0x6B, 0x00);
    VGACRWrite(pScrn, 0x6C, 0x00);

    ViaOutputsRestore(pScrn->scrnIndex);

    /* restore memory content */
    if (Regs->VGAFB && pVia->FBBase)
        memcpy(pVia->FBBase, Regs->VGAFB, 256*1024);
    else
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   "%s: No VGA FB stored or no FB mapped.\n", __func__);

    /* restore the standard vga regs */
    vgaHWRestore(pScrn, &VGAHWPTR(pScrn)->SavedReg, VGA_SR_MODE);

    /* restore extended regs */
    VGASRWrite(pScrn, 0x14, Regs->SR14);
    VGASRWrite(pScrn, 0x15, Regs->SR15);
    VGASRWrite(pScrn, 0x16, Regs->SR16);
    VGASRWrite(pScrn, 0x17, Regs->SR17);
    VGASRWrite(pScrn, 0x18, Regs->SR18);
    VGASRWrite(pScrn, 0x19, Regs->SR19);
    VGASRWrite(pScrn, 0x1A, Regs->SR1A);
    VGASRWrite(pScrn, 0x1B, Regs->SR1B);
    VGASRWrite(pScrn, 0x1C, Regs->SR1C);
    VGASRWrite(pScrn, 0x1D, Regs->SR1D);
    VGASRWrite(pScrn, 0x1E, Regs->SR1E);
    VGASRWrite(pScrn, 0x1F, Regs->SR1F);

    VGASRWrite(pScrn, 0x22, Regs->SR22);
    VGASRWrite(pScrn, 0x23, Regs->SR23);
    VGASRWrite(pScrn, 0x24, Regs->SR24);
    VGASRWrite(pScrn, 0x25, Regs->SR25);
    VGASRWrite(pScrn, 0x26, Regs->SR26);
    VGASRWrite(pScrn, 0x27, Regs->SR27);
    VGASRWrite(pScrn, 0x28, Regs->SR28);
    VGASRWrite(pScrn, 0x29, Regs->SR29);
    VGASRWrite(pScrn, 0x2A, Regs->SR2A);
    VGASRWrite(pScrn, 0x2B, Regs->SR2B);

    VGASRWrite(pScrn, 0x2E, Regs->SR2E);

    VGASRWrite(pScrn, 0x44, Regs->SR44);
    VGASRWrite(pScrn, 0x45, Regs->SR45);
    VGASRWrite(pScrn, 0x46, Regs->SR46);
    VGASRWrite(pScrn, 0x47, Regs->SR47);
    VGASRWrite(pScrn, 0x4A, Regs->SR4A);
    VGASRWrite(pScrn, 0x4B, Regs->SR4B);
    VGASRWrite(pScrn, 0x4C, Regs->SR4C);

    /* reset dotclocks */
    VGASRMask(pScrn, 0x40, 0x06, 0x06);
    VGASRMask(pScrn, 0x40, 0x00, 0x06);

    VGACRWrite(pScrn, 0x13, Regs->CR13);
    VGACRWrite(pScrn, 0x32, Regs->CR32);
    VGACRWrite(pScrn, 0x33, Regs->CR33);
    VGACRWrite(pScrn, 0x34, Regs->CR34);
    VGACRWrite(pScrn, 0x35, Regs->CR35);
    VGACRWrite(pScrn, 0x36, Regs->CR36);

    /* Restore LCD control regs */
    for (i = 0; i < 68; i++)
        VGACRWrite(pScrn, i + 0x50, Regs->CRTCRegs[i]);

    switch (pVia->Chipset) {
    case VT3108:
    case VT3344:
        VGACRWrite(pScrn, 0x97, Regs->CR97);
        VGACRWrite(pScrn, 0x99, Regs->CR99);
	break;
    default:
	break;
    }

    /* Re-enable Panel before returning */
    if (Regs->CRTCRegs[0x1A] & 0x08) { /* CR6A: LCD Enabled? */
        struct ViaOutput *Output = pVia->Outputs;

        while (Output) {
            if ((Output->Type == OUTPUT_PANEL) && Output->Power)
                Output->Power(Output, TRUE);

            Output = Output->Next;
        }
    }

    /* Reset clock */
    tmp = VGAMiscRead(pScrn);
    VGAMiscWrite(pScrn, tmp);

    vgaHWProtect(pScrn, FALSE);
    VGACRMask(pScrn, 0x11, 0x80, 0x80);
}

/*
 *
 */
static void
ViaMMIOEnable(ScrnInfoPtr pScrn)
{
    VGAEnableMask(pScrn, 0x01, 0x01);

    VGAMiscMask(pScrn, 0x01, 0x01); /* Colour! */

    /* Unlock Extended IO Space */
    VGASRWrite(pScrn, 0x10, 0x01);

    /* Enable MMIO */
    /* VT1122 crashes on VT switch, find out what other part is involved
       there for the proper solution. */
    if (VIAPTR(pScrn)->Chipset != VT1122)
	VGASRMask(pScrn, 0x1A, 0x02, 0x02);

    /* set PCI burst to 1 wait state */
    /* TODO: move me to base hw init */
    VGASRMask(pScrn, 0x1A, 0x04, 0x04);

    VGACRMask(pScrn, 0x11, 0x00, 0x80); /* modify starting address */
    VGACRMask(pScrn, 0x03, 0x80, 0x80); /* enable vertical retrace access */
    VGACRMask(pScrn, 0x47, 0x00, 0x01); /* unlock CRT registers */
}

/*
 *
 */
static Bool
VIAMapMMIO(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);

    VIAFUNC(pScrn);

#ifdef XSERVER_LIBPCIACCESS
    pVia->MmioBase = pVia->PciInfo->regions[1].base_addr;
#else
    pVia->MmioBase = pVia->PciInfo->memBase[1];
#endif

    xf86DrvMsg(pScrn->scrnIndex, X_PROBED,
               "mapping MMIO @ 0x%lx with size 0x%x\n",
               pVia->MmioBase, VIA_MMIO_REGSIZE);

#ifdef XSERVER_LIBPCIACCESS
    if (pci_device_map_range(pVia->PciInfo, pVia->MmioBase, VIA_MMIO_REGSIZE,
			     PCI_DEV_MAP_FLAG_WRITABLE, (pointer *) &pVia->MapBase))
	pVia->MapBase = NULL;
#else
    pVia->MapBase = xf86MapPciMem(pScrn->scrnIndex, VIDMEM_MMIO, pVia->PciTag,
                                  pVia->MmioBase, VIA_MMIO_REGSIZE);
#endif

    if (!pVia->MapBase) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   "%s: Unable to map IO.\n", __func__);
        return FALSE;
    }

    xf86DrvMsg(pScrn->scrnIndex, X_PROBED,
               "mapping BitBlt MMIO @ 0x%lx with size 0x%x\n",
               pVia->MmioBase + VIA_MMIO_BLTBASE, VIA_MMIO_BLTSIZE);

#ifdef XSERVER_LIBPCIACCESS
    if (pci_device_map_range(pVia->PciInfo, pVia->MmioBase + VIA_MMIO_BLTBASE,
			     VIA_MMIO_BLTSIZE, PCI_DEV_MAP_FLAG_WRITABLE,
			     (pointer *) &pVia->BltBase))
	pVia->BltBase = NULL;
#else
    pVia->BltBase = xf86MapPciMem(pScrn->scrnIndex, VIDMEM_MMIO, pVia->PciTag,
                                  pVia->MmioBase + VIA_MMIO_BLTBASE,
                                  VIA_MMIO_BLTSIZE);
#endif
    if (!pVia->BltBase) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   "%s: Unable to map BitBlt IO area.\n", __func__);
        return FALSE;
    }

    vgaHWSetMmioFuncs(VGAHWPTR(pScrn), pVia->MapBase, 0x8000);
    VGAHWPTR(pScrn)->IOBase = 0x3D0;

    return TRUE;
}

/*
 * Gets called at MapFB (ScreenInit) and EnterVTs
 */
static void
ViaFBEnable(ScrnInfoPtr pScrn)
{
    /* Enable writing to all VGA memory planes */
    VGASRWrite(pScrn, 0x02, 0x0F);

    /* Enable Extended VGA memory */
    VGASRWrite(pScrn, 0x04, 0x0E);

    /* Enable Extended Memory access */
    VGASRMask(pScrn, 0x1A, 0x08, 0x08);
}

#ifdef XSERVER_LIBPCIACCESS

/*
 * Bending over backwards libpciaccess.
 */
static void
ViaMapFBBase(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);
    struct pci_device *dev = pVia->PciInfo;
    pciaddr_t oldbase, oldsize;
    int ret;

    VIAFUNC(pScrn);

    oldbase = dev->regions[0].base_addr;
    oldsize = dev->regions[0].size;

    /* fool libpciaccess */
    if (pVia->FrameBufferBase != oldbase) {
	dev->regions[0].base_addr = pVia->FrameBufferBase;
	dev->regions[0].size = pScrn->videoRam << 10;
    }

    ret = pci_device_map_range(dev, pVia->FrameBufferBase,
			       pScrn->videoRam << 10,
			       PCI_DEV_MAP_FLAG_WRITABLE |
			       PCI_DEV_MAP_FLAG_WRITE_COMBINE,
			       (pointer *) &pVia->FBBase);
    if (ret) {
	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
		   "%s: Could not map framebuffer: %d\n", __func__, ret);
	pVia->FBBase = NULL;
    }

    if (pVia->FrameBufferBase != oldbase) {
	dev->regions[0].base_addr = oldbase;
	dev->regions[0].size = oldsize;
    }
}

#else

/*
 * FIXME: This is a hack to get rid of offending wrongly sized
 * MTRR regions set up by the VIA BIOS. Should be taken care of
 * in the OS support layer.
 */
static void
ViaBrokenMTRRHack(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);
    CARD8 *tmp;

    VIAFUNC(pScrn);

    tmp = xf86MapPciMem(pScrn->scrnIndex, VIDMEM_MMIO, pVia->PciTag,
                        pVia->FrameBufferBase, pScrn->videoRam << 10 );
    xf86UnMapVidMem(pScrn->scrnIndex, (pointer)tmp, pScrn->videoRam << 10);

    /* And, as if this wasn't enough, 2.6 series kernels doesn't remove MTRR
     * regions on the first attempt. Try again. */
    tmp = xf86MapPciMem(pScrn->scrnIndex, VIDMEM_MMIO, pVia->PciTag,
                        pVia->FrameBufferBase, pScrn->videoRam << 10);
    xf86UnMapVidMem(pScrn->scrnIndex, (pointer)tmp, pScrn->videoRam << 10);
}

/*
 * The good old days.
 */
static void
ViaMapFBBase(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);

    VIAFUNC(pScrn);

    ViaBrokenMTRRHack(pScrn);
    pVia->FBBase = xf86MapPciMem(pScrn->scrnIndex, VIDMEM_FRAMEBUFFER,
                                 pVia->PciTag, pVia->FrameBufferBase,
                                 pScrn->videoRam << 10);
}
#endif /* XSERVER_LIBPCIACCESS */


/*
 *
 */
static Bool
ViaMapFB(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);

    VIAFUNC(pScrn);

    /* we're going to try to map before ScreenInit, because we need direct access
       during store/restore. */
    if (pVia->FBBase)
        return TRUE;

    ViaFBBaseGet(pScrn);
    xf86DrvMsg(pScrn->scrnIndex, X_PROBED,
               "mapping framebuffer @ 0x%lx with size 0x%lx\n",
               pVia->FrameBufferBase, (long unsigned int) (pScrn->videoRam << 10));

    ViaMapFBBase(pScrn);

    if (!pVia->FBBase) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   "Internal error: could not map framebuffer\n");
        return FALSE;
    }

    ViaDebug(pScrn->scrnIndex, "%s: FBBase: %p\n", __func__, pVia->FBBase);

    pScrn->memPhysBase = pVia->FrameBufferBase;
    pScrn->fbOffset = 0;

    ViaFBEnable(pScrn);

    return TRUE;
}


static void
VIAUnmapMem(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);

    VIAFUNC(pScrn);

    /* Disable MMIO */
    VGASRMask(pScrn, 0x1A, 0x00, 0x02);

    if (pVia->MapBase)
#ifdef XSERVER_LIBPCIACCESS
	pci_device_unmap_range(pVia->PciInfo, (pointer)pVia->MapBase, VIA_MMIO_REGSIZE);
#else
        xf86UnMapVidMem(pScrn->scrnIndex, (pointer)pVia->MapBase, VIA_MMIO_REGSIZE);
#endif

    if (pVia->BltBase)
#ifdef XSERVER_LIBPCIACCESS
	pci_device_unmap_range(pVia->PciInfo, (pointer)pVia->BltBase, VIA_MMIO_BLTSIZE);
#else
        xf86UnMapVidMem(pScrn->scrnIndex, (pointer)pVia->BltBase, VIA_MMIO_BLTSIZE);
#endif

    if (pVia->FBBase)
#ifdef XSERVER_LIBPCIACCESS
	pci_device_unmap_range(pVia->PciInfo, (pointer)pVia->FBBase, pScrn->videoRam << 10);
#else
        xf86UnMapVidMem(pScrn->scrnIndex, (pointer)pVia->FBBase, pScrn->videoRam << 10);
#endif
    pVia->FBBase = 0;
}

/*
 * Doesn't do secondary/merged currently.
 */
static void
ViaPaletteLoad(ScrnInfoPtr pScrn, int numColors, int *indices,
               LOCO *colors, VisualPtr pVisual)
{
    VIAPtr pVia = VIAPTR(pScrn);

    VIAFUNC(pScrn);

    if ((pVia->Crtc[0]->scrnIndex == pScrn->scrnIndex) && pVia->Crtc[0]->Active)
        pVia->Crtc[0]->GammaSet(pVia->Crtc[0], numColors, indices, colors);

    if ((pVia->Crtc[1]->scrnIndex == pScrn->scrnIndex) && pVia->Crtc[1]->Active)
        pVia->Crtc[1]->GammaSet(pVia->Crtc[1], numColors, indices, colors);
}

/*
 * standard VGA Overscan register.
 */
static void
ViaOverScan(ScrnInfoPtr pScrn, int overscan)
{
    VIAFUNC(pScrn);

    switch (pScrn->bitsPerPixel) {
    case 16:
        VGAARWrite(pScrn, 0x11, overscan * 4);
        break;
    case 8:
    case 24:
    case 32:
        VGAARWrite(pScrn, 0x11, overscan);
        break;
    default:
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "%s: Unsupported bitdepth: %d\n",
                   __func__, pScrn->bitsPerPixel);
        break;
    }
}

/*
 *
 */
static Bool
ViaWrapvgaHWSaveScreen(ScreenPtr pScreen, int mode)
{
    return vgaHWSaveScreen(pScreen, mode);
}

/*
 *
 */
static Bool
VIAScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
{
    ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
    VIAPtr pVia = VIAPTR(pScrn);

    VIAFUNC(pScrn);

    if (!ViaPciDeviceEnable(pVia)) {
	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
		   "%s: Failed to enable pci device.\n", __func__);
	return FALSE;
    }

    if (!VIAMapMMIO(pScrn))
        return FALSE;

    ViaMMIOEnable(pScrn);

    VIASave(pScrn);

    if (!ViaMapFB(pScrn))
        return FALSE;

    /* Stop previous content from disturbing users.
     * Don't do this from MapFB itself, that ruins text mode restoration. */
    memset(VIAPTR(pScrn)->FBBase, 0x00,
           pScrn->displayWidth * pScrn->virtualY * (pScrn->bitsPerPixel >> 3));

    vgaHWSaveScreen(pScreen, SCREEN_SAVER_ON);

    ViaCrtcModeSetInitial(pScrn);

    if (pVia->Crtc[0]->Active)
        ViaOutputsPower(pVia->Crtc[0], FALSE);
    if (pVia->Crtc[1]->Active)
        ViaOutputsPower(pVia->Crtc[1], FALSE);

#ifdef HAVE_DRI
    /* Hack: need to figure out proper ordering for WriteMode,
       3DInit and DRI Init */
    pVia->directRenderingEnabled = pVia->UseDRI;
#endif

    VIAWriteMode(pScrn, pScrn->currentMode);

    pScrn->AdjustFrame(scrnIndex, pScrn->frameX0, pScrn->frameY0, 0);

    ViaDebug(scrnIndex, "- Blanked\n");

    miClearVisualTypes();

    /* IsSecondary took the 8bpp route here too */
    if (pScrn->bitsPerPixel > 8) {
        if (!miSetVisualTypes(pScrn->depth, TrueColorMask,
                              pScrn->rgbBits, pScrn->defaultVisual))
            return FALSE;
        if (!miSetPixmapDepths())
            return FALSE;
    } else {
        if (!miSetVisualTypes(pScrn->depth, miGetDefaultVisualMask(pScrn->depth),
                              pScrn->rgbBits, pScrn->defaultVisual))
            return FALSE;
        if (!miSetPixmapDepths())
            return FALSE;
    }

#ifdef HAVE_DRI
    if (pVia->UseDRI) {
	pVia->directRenderingEnabled = VIADRIScreenInit(pScrn, pScreen);
	if (!pVia->directRenderingEnabled) {
	    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "[drm] Not using DMA for "
		       "copying to FB (no DRI).\n");
	    pVia->UseDMACopy = FALSE;
	}
    } else
	pVia->directRenderingEnabled = FALSE;
#endif

    ViaDebug(scrnIndex, "- Visuals set up\n");

    if (!VIAInternalScreenInit(pScrn, pScreen))
	return FALSE;

    xf86SetBlackWhitePixels(pScreen);
    ViaDebug(scrnIndex, "- B & W\n");

    if (pScrn->bitsPerPixel > 8) {
        VisualPtr visual;

        visual = pScreen->visuals + pScreen->numVisuals;
        while (--visual >= pScreen->visuals) {
            if ((visual->class | DynamicClass) == DirectColor) {
                visual->offsetRed = pScrn->offset.red;
                visual->offsetGreen = pScrn->offset.green;
                visual->offsetBlue = pScrn->offset.blue;
                visual->redMask = pScrn->mask.red;
                visual->greenMask = pScrn->mask.green;
                visual->blueMask = pScrn->mask.blue;
            }
        }
    }

#ifdef USE_FB
    /* must be after RGB ordering fixed */
    fbPictureInit(pScreen, 0, 0);
#endif

    /* Take 3D engine out of reset, also needed for VQ */
    VGASRMask(pScrn, 0x1A, 0x00, 0x40);

    if (pVia->AccelMethod == VIA_ACCEL_XAA)
	VIAInitAccel(pScrn, pScreen);

    {
	unsigned int BytesPerLine =
	    pScrn->displayWidth * (pScrn->bitsPerPixel >> 3);
	BoxRec AvailFBArea;

	AvailFBArea.x1 = pVia->FBScanoutStart % BytesPerLine;
	AvailFBArea.y1 = pVia->FBScanoutStart / BytesPerLine;
	AvailFBArea.x2 = AvailFBArea.x1 + pScrn->displayWidth;
	AvailFBArea.y2 = AvailFBArea.y1 +
	    ((pVia->FBScanoutSize + pVia->FBOffscreenSize) / BytesPerLine);

	ViaDebug(pScrn->scrnIndex, "Initialising FB Manager for (%4d, %4d) -> (%4d, %4d)\n",
		 AvailFBArea.x1, AvailFBArea.y1, AvailFBArea.x2, AvailFBArea.y2);

	xf86InitFBManager(pScreen, &AvailFBArea);
    }

    miInitializeBackingStore(pScreen);
    xf86SetBackingStore(pScreen);
    /*xf86SetSilkenMouse(pScreen);*/
    ViaDebug(scrnIndex, "- Backing store set up\n");

    miDCInitialize(pScreen, xf86GetPointerScreenFuncs());
    ViaDebug(scrnIndex, "- SW cursor set up\n");

    if (pVia->hwcursor) {
	pVia->Cursor = VIACursorInit(pScrn, pScreen);
	if (!pVia->Cursor)
	    xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
		       "Hardware cursor initialization failed\n");
    }

    if (pVia->AccelMethod == VIA_ACCEL_SHADOWFB)
	ViaShadowFBInit(pScrn, pScreen);

    if (!miCreateDefColormap(pScreen))
        return FALSE;

    if (!xf86HandleColormaps(pScreen, 256, 8, ViaPaletteLoad, ViaOverScan,
                             CMAP_RELOAD_ON_MODE_SWITCH
                             | CMAP_PALETTED_TRUECOLOR))
        return FALSE;

    vgaHWBlankScreen(pScrn, TRUE);

    pVia->CloseScreen = pScreen->CloseScreen;
    pScreen->SaveScreen = ViaWrapvgaHWSaveScreen;
    pScreen->CloseScreen = VIACloseScreen;

    xf86DPMSInit(pScreen, VIADPMS, 0);
    ViaDebug(scrnIndex, "- DPMS set up\n");

    ViaDebug(scrnIndex, "- Color maps etc. set up\n");

#ifdef HAVE_DRI
    if (pVia->directRenderingEnabled)
	pVia->directRenderingEnabled = VIADRIFinishScreenInit(pScrn, pScreen);

    if (pVia->directRenderingEnabled)
        xf86DrvMsg(scrnIndex, X_INFO, "direct rendering enabled\n");
    else {
        xf86DrvMsg(scrnIndex, X_INFO, "direct rendering disabled\n");
	ViaMemInit(pScrn, pScreen);
    }
#else
    ViaMemInit(pScrn, pScreen);
#endif

#ifdef _VIA_VIDEO_H_
    if (pVia->UseXv)
        ViaVideoInit(pScrn, pScreen);
#endif

    if (pVia->Crtc[0]->Active)
        ViaOutputsPower(pVia->Crtc[0], TRUE);
    if (pVia->Crtc[1]->Active)
        ViaOutputsPower(pVia->Crtc[1], TRUE);

    if (serverGeneration == 1)
        xf86ShowUnusedOptions(pScrn->scrnIndex, pScrn->options);

    ViaDebug(scrnIndex, "- Done\n");
    return TRUE;
}

static int
VIAInternalScreenInit(ScrnInfoPtr pScrn, ScreenPtr pScreen)
{
    VIAPtr          pVia = VIAPTR(pScrn);
    int             width, height, displayWidth;
    unsigned char   *FBStart;

    VIAFUNC(pScrn);

    if (pVia->rotate) {
	height = pScrn->virtualX;
	width = pScrn->virtualY;
    } else {
	width = pScrn->virtualX;
	height = pScrn->virtualY;
    }

    if (pVia->AccelMethod == VIA_ACCEL_SHADOWFB) {
	pVia->ShadowPitch = BitmapBytePad(pScrn->bitsPerPixel * width);
	pVia->ShadowPtr = xnfcalloc(1, pVia->ShadowPitch * height);
	displayWidth = pVia->ShadowPitch / (pScrn->bitsPerPixel >> 3);
	FBStart = pVia->ShadowPtr;
    } else {
	pVia->ShadowPtr = NULL;
	displayWidth = pScrn->displayWidth;
	FBStart = pVia->FBBase;
    }

#ifdef USE_FB
    return fbScreenInit(pScreen, FBStart, width, height,
			pScrn->xDpi, pScrn->yDpi, displayWidth,
			pScrn->bitsPerPixel);
#else
    switch (pScrn->bitsPerPixel) {
    case 8:
	return cfbScreenInit(pScreen, FBStart, width, height, pScrn->xDpi,
			     pScrn->yDpi, displayWidth);
    case 16:
	return cfb16ScreenInit(Screen, FBStart, width, height, pScrn->xDpi,
			       pScrn->yDpi, displayWidth);
    case 32:
	return cfb32ScreenInit(pScreen, FBStart, width, height, pScrn->xDpi,
			       pScrn->yDpi, displayWidth);
    default:
	xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Internal error: invalid bpp (%d) in "
		   "%s\n", pScrn->bitsPerPixel, __func__);
	return FALSE;
    }
#endif
}

static void
VIAWriteMode(ScrnInfoPtr pScrn, DisplayModePtr mode)
{
    VIAPtr pVia = VIAPTR(pScrn);

    VIAFUNC(pScrn);

    /* this is waaaay too simplistic and even plain wrong. */
#ifdef _VIA_VIDEO_H_
    if (pVia->Swov)
        pVia->Swov->Allowed = FALSE;
#endif

    pScrn->vtSema = TRUE;

    /* FIXME - need DRI lock for this bit - see i810 */
    ViaModeSet(pScrn, mode);

    /* Enable the graphics engine. */
    if (pVia->AccelInfoRec)
	VIAInitialize2DEngine(pScrn);

#ifdef HAVE_DRI
    if (pVia->directRenderingEnabled)
	VIAInitialize3DEngine(pScrn);
#endif

    VIAAdjustFrame(pScrn->scrnIndex, pScrn->frameX0, pScrn->frameY0, 0);
}


static Bool
VIACloseScreen(int scrnIndex, ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
    VIAPtr pVia = VIAPTR(pScrn);

    VIAFUNC(pScrn);

    /* Is the display currently visible ? */
    if (pScrn->vtSema) {
#ifdef HAVE_DRI
	if (pVia->directRenderingEnabled)
	    DRILock(pScreen, 0);
#endif
	/* Wait Hardware Engine idle to exit graphical mode */
	VIAAccelSync(pScrn);

	VIACursorHide(pVia);

	/* Put 3D engine in reset, includes VQ */
	VGASRMask(pScrn, 0x1A, 0x40, 0x40);

	if (pVia->VQEnable)
	    ViaVQDisable(pScrn);
    }

#ifdef HAVE_DRI
    if (pVia->directRenderingEnabled)
	VIADRICloseScreen(pScrn, pScreen);
#endif

#ifdef _VIA_VIDEO_H_
    if (pVia->Swov)
        ViaVideoDestroy(pScrn);
#endif

    if (pVia->AccelInfoRec) {
        XAADestroyInfoRec(pVia->AccelInfoRec);
        pVia->AccelInfoRec = NULL;
    }

    if (pVia->ShadowPtr) {
        free(pVia->ShadowPtr);
        pVia->ShadowPtr = NULL;
    }

    if (pScrn->vtSema) {
        VIARestore(pScrn);
        VIAUnmapMem(pScrn);
    }

    ViaPciDeviceDisable(pVia);

    pScrn->vtSema = FALSE;
    pScreen->CloseScreen = pVia->CloseScreen;
    return (*pScreen->CloseScreen)(scrnIndex, pScreen);
}

/*
 * This only gets called when a screen is being deleted.  It does not
 * get called routinely at the end of a server generation.
 */
static void VIAFreeScreen(int scrnIndex, int flags)
{
    VIAFUNCI(scrnIndex);

    VIAFreeRec(xf86Screens[scrnIndex]);

    if (xf86LoaderCheckSymbol("vgaHWFreeHWRec"))
        vgaHWFreeHWRec(xf86Screens[scrnIndex]);
}

static void
VIAAdjustFrame(int scrnIndex, int x, int y, int flags)
{
    VIAPtr pVia = VIAPTR(xf86Screens[scrnIndex]);

    if ((pVia->Crtc[0]->scrnIndex == scrnIndex) && pVia->Crtc[0]->Active)
        pVia->Crtc[0]->FrameSet(pVia->Crtc[0], x, y);

    if ((pVia->Crtc[1]->scrnIndex == scrnIndex) && pVia->Crtc[1]->Active)
        pVia->Crtc[1]->FrameSet(pVia->Crtc[1], x, y);
}


static Bool
VIASwitchMode(int scrnIndex, DisplayModePtr mode, int flags)
{
    ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
    VIAPtr      pVia = VIAPTR(pScrn);

    VIAFUNC(pScrn);

    if (pVia->Crtc[0]->Active)
        ViaOutputsPower(pVia->Crtc[0], FALSE);
    if (pVia->Crtc[1]->Active)
        ViaOutputsPower(pVia->Crtc[1], FALSE);

#ifdef HAVE_DRI
    if (pVia->directRenderingEnabled)
	DRILock(pScrn->pScreen, 0);
#endif

    VIAAccelSync(pScrn);

#ifdef HAVE_DRI
    if (pVia->directRenderingEnabled && pVia->CommandBufferActive)
	VIADRMCommandBufferStop(pVia);
#endif

    if (pVia->VQEnable)
	ViaVQDisable(pScrn);

    VIAWriteMode(pScrn, mode);

    if (pVia->VQEnable)
	ViaVQEnable(pScrn);

#ifdef HAVE_DRI
    if (pVia->directRenderingEnabled) {
	if (pVia->UseCommandBuffer)
	    VIADRMCommandBufferStart(pVia);
	DRIUnlock(screenInfo.screens[scrnIndex]);
    }
#endif

    if (pVia->Crtc[0]->Active)
        ViaOutputsPower(pVia->Crtc[0], TRUE);
    if (pVia->Crtc[1]->Active)
        ViaOutputsPower(pVia->Crtc[1], TRUE);

    return TRUE;
}

/*
 *
 */
static void VIADPMS(ScrnInfoPtr pScrn, int mode, int flags)
{
    VIAPtr pVia = VIAPTR(pScrn);

    VIAFUNC(pScrn);

    switch (mode) {
    case DPMSModeOn:
        /* X's DPMS calls vgaHWBlankScreen while it has no business there */
        VGASRMask(pScrn, 0x01, 0x00, 0x20);
        if (pVia->Crtc[0]->Active)
            ViaOutputsPower(pVia->Crtc[0], TRUE);
        if (pVia->Crtc[1]->Active)
            ViaOutputsPower(pVia->Crtc[1], TRUE);
        break;
    case DPMSModeStandby:
    case DPMSModeSuspend:
    case DPMSModeOff:
        if (pVia->Crtc[0]->Active)
            ViaOutputsPower(pVia->Crtc[0], FALSE);
        if (pVia->Crtc[1]->Active)
            ViaOutputsPower(pVia->Crtc[1], FALSE);
        break;
    default:
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Invalid DPMS mode %d\n", mode);
        break;
    }
    return;
}

#ifdef HAVE_DRI
void
VIAInitialize3DEngine(ScrnInfoPtr pScrn)
{
    VIAPtr  pVia = VIAPTR(pScrn);
    int i;

    VIAFUNC(pScrn);

    /*
     * FIXME: Reinitializing the 3D engine after VT switch seems to be the way
     * to go to avoid hangs and rendering errors. But what happens with secondary?
     * Test this for a while, but keep the old code in case...
     */

    if (1) {

        VIASETREG(0x43C, 0x00010000);

        for (i = 0; i <= 0x7D; i++)
        {
            VIASETREG(0x440, (CARD32) i << 24);
        }

        VIASETREG(0x43C, 0x00020000);

        for (i = 0; i <= 0x94; i++)
        {
            VIASETREG(0x440, (CARD32) i << 24);
        }

        VIASETREG(0x440, 0x82400000);

        VIASETREG(0x43C, 0x01020000);


        for (i = 0; i <= 0x94; i++)
        {
            VIASETREG(0x440, (CARD32) i << 24);
        }

        VIASETREG(0x440, 0x82400000);
        VIASETREG(0x43C, 0xfe020000);

        for (i = 0; i <= 0x03; i++)
        {
            VIASETREG(0x440, (CARD32) i << 24);
        }

        VIASETREG(0x43C, 0x00030000);

        for (i = 0; i <= 0xff; i++)
        {
            VIASETREG(0x440, 0);
        }
        VIASETREG(0x43C, 0x00100000);
        VIASETREG(0x440, 0x00333004);
        VIASETREG(0x440, 0x10000002);
        VIASETREG(0x440, 0x60000000);
        VIASETREG(0x440, 0x61000000);
        VIASETREG(0x440, 0x62000000);
        VIASETREG(0x440, 0x63000000);
        VIASETREG(0x440, 0x64000000);

        VIASETREG(0x43C, VIA_REG_VIRTUAL_QUEUE);

        if ((pVia->Chipset == VT3122) && (pVia->HostRev >= 3))
            VIASETREG(0x440,0x40008c0f);
        else
            VIASETREG(0x440,0x4000800f);

        VIASETREG(0x440,0x44000000);
        VIASETREG(0x440,0x45080C04);
        VIASETREG(0x440,0x46800408);
        VIASETREG(0x440,0x50000000); /* WTF!!! */
        VIASETREG(0x440,0x51000000); /* Why are we killing VQ sizing? */
        VIASETREG(0x440,0x52000000);
        VIASETREG(0x440,0x53000000);

        xf86DrvMsg(pScrn->scrnIndex, X_INFO,
            "3D Engine has been initialized.\n");
    }

    VIASETREG(0x43C,VIA_REG_VIRTUAL_QUEUE);
    VIASETREG(0x440,0x08000001);
    VIASETREG(0x440,0x0A000183);
    VIASETREG(0x440,0x0B00019F);
    VIASETREG(0x440,0x0C00018B);
    VIASETREG(0x440,0x0D00019B);
    VIASETREG(0x440,0x0E000000);
    VIASETREG(0x440,0x0F000000);
    VIASETREG(0x440,0x10000000);
    VIASETREG(0x440,0x11000000);
    VIASETREG(0x440,0x20000000);
}
#endif


/*
 * Since X lacks anything of the sort.
 * Used by VIAFUNC macro.
 */
void
ViaDebug(int scrnIndex, const char *format, ...)
{
    va_list ap;

    va_start(ap, format);
    xf86VDrvMsgVerb(scrnIndex, X_INFO, LOG_DEBUG, format, ap);
    va_end(ap);

}
