/*
 * Copyright 2004-2009 Luc Verhaegen.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_XF86_ANSIC_H
#include "xf86_ansic.h"
#endif

#include "compiler.h"
#include "via_driver.h"
#include "via_regs.h"

/*
 * We are kind of reinventing the wheel here, but i prefer better readable
 * and uniform code, and we never do PIO accesses anyway.
 */

#define IO(scrnIndex) VIAPTR(xf86Screens[(scrnIndex)])->MapBase
#define VGA(offset) (0x8000 + (offset))

#define VGA_SR_INDEX VGA(0x3C4)
#define VGA_SR_DATA  VGA(0x3C5)

CARD8
VGASRReadI(int scrnIndex, CARD8 index)
{
    void *io = IO(scrnIndex);

    MMIO_OUT8(io, VGA_SR_INDEX, index);
    return MMIO_IN8(io, VGA_SR_DATA);
}

void
VGASRWriteI(int scrnIndex, CARD8 index, CARD8 data)
{
    void *io = IO(scrnIndex);

    MMIO_OUT8(io, VGA_SR_INDEX, index);
    MMIO_OUT8(io, VGA_SR_DATA, data);
}

void
VGASRMaskI(int scrnIndex, CARD8 index, CARD8 data, CARD8 mask)
{
    void *io = IO(scrnIndex);
    CARD8 tmp;

    MMIO_OUT8(io, VGA_SR_INDEX, index);

    tmp = MMIO_IN8(io, VGA_SR_DATA);

    tmp &= ~mask;
    tmp |= (data & mask);

    MMIO_OUT8(io, VGA_SR_DATA, tmp);
}

#define VGA_CR_INDEX VGA(0x3D4)
#define VGA_CR_DATA  VGA(0x3D5)

CARD8
VGACRReadI(int scrnIndex, CARD8 index)
{
    void *io = IO(scrnIndex);

    MMIO_OUT8(io, VGA_CR_INDEX, index);
    return MMIO_IN8(io, VGA_CR_DATA);
}

void
VGACRWriteI(int scrnIndex, CARD8 index, CARD8 data)
{
    void *io = IO(scrnIndex);

    MMIO_OUT8(io, VGA_CR_INDEX, index);
    MMIO_OUT8(io, VGA_CR_DATA, data);
}

void
VGACRMaskI(int scrnIndex, CARD8 index, CARD8 data, CARD8 mask)
{
    void *io = IO(scrnIndex);
    CARD8 tmp;

    MMIO_OUT8(io, VGA_CR_INDEX, index);

    tmp = MMIO_IN8(io, VGA_CR_DATA);

    tmp &= ~mask;
    tmp |= (data & mask);

    MMIO_OUT8(io, VGA_CR_DATA, tmp);
}

#define VGA_GR_INDEX VGA(0x3CE)
#define VGA_GR_DATA  VGA(0x3CF)

CARD8
VGAGRReadI(int scrnIndex, CARD8 index)
{
    void *io = IO(scrnIndex);

    MMIO_OUT8(io, VGA_GR_INDEX, index);
    return MMIO_IN8(io, VGA_GR_DATA);
}

void
VGAGRWriteI(int scrnIndex, CARD8 index, CARD8 data)
{
    void *io = IO(scrnIndex);

    MMIO_OUT8(io, VGA_GR_INDEX, index);
    MMIO_OUT8(io, VGA_GR_DATA, data);
}

void
VGAGRMaskI(int scrnIndex, CARD8 index, CARD8 data, CARD8 mask)
{
    void *io = IO(scrnIndex);
    CARD8 tmp;

    MMIO_OUT8(io, VGA_GR_INDEX, index);

    tmp = MMIO_IN8(io, VGA_GR_DATA);

    tmp &= ~mask;
    tmp |= (data & mask);

    MMIO_OUT8(io, VGA_GR_DATA, tmp);
}

#define VGA_ENABLE VGA(0x3C3)

CARD8
VGAEnableReadI(int scrnIndex)
{
    void *io = IO(scrnIndex);
    return MMIO_IN8(io, VGA_ENABLE);
}

void
VGAEnableWriteI(int scrnIndex, CARD8 data)
{
    void *io = IO(scrnIndex);
    MMIO_OUT8(io, VGA_ENABLE, data);
}

void
VGAEnableMaskI(int scrnIndex, CARD8 data, CARD8 mask)
{
    void *io = IO(scrnIndex);
    CARD8 tmp;

    tmp = MMIO_IN8(io, VGA_ENABLE);

    tmp &= ~mask;
    tmp |= (data & mask);

    MMIO_OUT8(io, VGA_ENABLE, tmp);
}

#define VGA_MISC_READ VGA(0x3CC)
#define VGA_MISC_WRITE VGA(0x3C2)

CARD8
VGAMiscReadI(int scrnIndex)
{
    void *io = IO(scrnIndex);
    return MMIO_IN8(io, VGA_MISC_READ);
}

void
VGAMiscWriteI(int scrnIndex, CARD8 data)
{
    void *io = IO(scrnIndex);
    MMIO_OUT8(io, VGA_MISC_WRITE, data);
}

void
VGAMiscMaskI(int scrnIndex, CARD8 data, CARD8 mask)
{
    void *io = IO(scrnIndex);
    CARD8 tmp;

    tmp = MMIO_IN8(io, VGA_MISC_READ);

    tmp &= ~mask;
    tmp |= (data & mask);

    MMIO_OUT8(io, VGA_MISC_WRITE, tmp);
}

#define VGA_STATUS0 VGA(0x3C2)

CARD8
VGAStatus0ReadI(int scrnIndex)
{
    void *io = IO(scrnIndex);
    return MMIO_IN8(io, VGA_STATUS0);
}

#define VGA_STATUS1 VGA(0x3DA)

CARD8
VGAStatus1ReadI(int scrnIndex)
{
    void *io = IO(scrnIndex);
    return MMIO_IN8(io, VGA_STATUS1);
}

#define VGA_DAC_MASK VGA(0x3C6)
#define VGA_DAC_RADDR VGA(0x3C7)
#define VGA_DAC_WADDR VGA(0x3C8)
#define VGA_DAC_DATA VGA(0x3C9)

CARD8
VGADACDataReadI(int scrnIndex)
{
    void *io = IO(scrnIndex);
    return MMIO_IN8(io, VGA_DAC_DATA);
}

void
VGADACDataWriteI(int scrnIndex, CARD8 data)
{
    void *io = IO(scrnIndex);
    MMIO_OUT8(io, VGA_DAC_DATA, data);
}

void
VGADACMaskWriteI(int scrnIndex, CARD8 data)
{
    void *io = IO(scrnIndex);
    MMIO_OUT8(io, VGA_DAC_MASK, data);
}

void
VGADACReadAddressWriteI(int scrnIndex, CARD8 data)
{
    void *io = IO(scrnIndex);
    MMIO_OUT8(io, VGA_DAC_RADDR, data);
}

void
VGADACWriteAddressWriteI(int scrnIndex, CARD8 data)
{
    void *io = IO(scrnIndex);
    MMIO_OUT8(io, VGA_DAC_WADDR, data);
}

#define VGA_AR_INDEX VGA(0x3C0)
#define VGA_AR_DATA_WRITE VGA(0x3C0)
#define VGA_AR_DATA_READ VGA(0x3C1)

CARD8
VGAARReadI(int scrnIndex, CARD8 index)
{
    void *io = IO(scrnIndex);

    MMIO_IN8(io, VGA_STATUS1);

    MMIO_OUT8(io, VGA_AR_INDEX, index | 0x20);
    return MMIO_IN8(io, VGA_AR_DATA_READ);
}

void
VGAARWriteI(int scrnIndex, CARD8 index, CARD8 data)
{
    void *io = IO(scrnIndex);

    MMIO_IN8(io, VGA_STATUS1);

    MMIO_OUT8(io, VGA_AR_INDEX, index | 0x20);
    MMIO_OUT8(io, VGA_AR_DATA_WRITE, data);
}

void
VGADACAccessEnableI(int scrnIndex)
{
    void *io = IO(scrnIndex);
    CARD8 tmp;

    MMIO_IN8(io, VGA_STATUS1);

    tmp = MMIO_IN8(io, VGA_AR_INDEX);
    MMIO_OUT8(io, VGA_AR_INDEX, tmp & ~0x20);
}

void
VGADACAccessDisableI(int scrnIndex)
{
    void *io = IO(scrnIndex);
    CARD8 tmp;

    MMIO_IN8(io, VGA_STATUS1);

    tmp = MMIO_IN8(io, VGA_AR_INDEX);
    MMIO_OUT8(io, VGA_AR_INDEX, tmp | 0x20);
}
