/*
 * Copyright 2008 Luc Verhaegen.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

/*
 * Wrapper driver that loads unichrome_drv.so.
 *
 * Compile with:
 * -D WRAPPER_NAME="bleh"
 * -D WRAPPER_NAME_REPLACE\(name\)=NAME_CAT\(bleh,name\)
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "xf86.h"

#if !defined(UNIXCPP) || defined(ANSICPP)
#define NAME_CAT(prefix,name) prefix##name
#else
#define NAME_CAT(prefix,name) prefix/**/name
#endif

/*
 *
 */
static XF86ModuleVersionInfo
WRAP_REPLACE(VersRec) = {
    WRAP_NAME,
    "http://unichrome.sf.net/",
    MODINFOSTRING1,
    MODINFOSTRING2,
#ifdef XORG_VERSION_CURRENT
    XORG_VERSION_CURRENT,
#else
    XF86_VERSION_CURRENT,
#endif
    0, 0, 0,
    ABI_CLASS_VIDEODRV,
    ABI_VIDEODRV_VERSION,
    MOD_CLASS_VIDEODRV,
    {0, 0, 0, 0}
};

/*
 *
 */
static pointer
WRAP_REPLACE(Setup)(
#ifdef HaveModuleDescPtr /* XFree86 4.6.0 */
                        ModuleDescPtr module,
#else
                        pointer module,
#endif
                        pointer options, int *errmaj, int *errmin)
{
    static int Initialised = FALSE;

    if (xf86DoConfigure && xf86DoConfigurePass1) {
	xf86Msg(X_WARNING, "%s: xserver is configuring, skipping wrapper"
		" driver.\n", __func__);

	if (errmaj)
            *errmaj = LDR_NOERROR;
        if (errmin)
	    *errmin = 0;

	return NULL;
    }

    if (Initialised) {
	if (errmaj)
            *errmaj = LDR_ONCEONLY;

	return NULL;
    }

    Initialised = TRUE;

    {
	GDevPtr *gdevs;
	int count, i;

	count = xf86MatchDevice(WRAP_NAME, &gdevs);

	if (!count) {
	    xf86Msg(X_ERROR, "%s: No devices found.\n", __func__);
	    return NULL;
	}

	/* change the name */
	for (i = 0; i < count; i++)
	    gdevs[i]->driver = "unichrome";

	free(gdevs);
    }

    xf86LoadOneModule("unichrome", options);

    return (pointer) 1;
}

/*
 *
 */
_X_EXPORT XF86ModuleData
WRAP_REPLACE(ModuleData) =
{
    &WRAP_REPLACE(VersRec),
    WRAP_REPLACE(Setup),
    NULL
};
